#ifndef _GFSFBSD_
#define _GFSFBSD_


/*  Include Files
    These should only be the ones necessary to compile the common code.  */

#include <sys/time.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/stat.h>
#include <lgfs.h>



/*  Memory allocation abstraction  */

#define gmalloc(x)  malloc((x), M_DEVBUF, M_WAITOK)
#define gfree(x,y)  free((x), M_DEVBUF)



/*  Memory copy abstraction */

#define memset(x,y,z) bzero(x,z)

#define GFS_IO_READ(b, c, f, l) uiomove((f), (l), (struct uio *)(b))
#define GFS_IO_READ_HOLE(b, c, l) { \
				      char *f = (char *)gmalloc((l)); \
				      memset(f, 0, (l)); \
				      GFS_IO_READ((b), (c), f, (l)); \
				      gfree(f, (l)); \
				  }
#define GFS_IO_WRITE(b, c, t, l) GFS_IO_READ((b), (c), (t), (l))



/*  printk() abstraction  */

#define KERN_EMERG      "<0>"   /* system is unusable                   */
#define KERN_ALERT      "<1>"   /* action must be taken immediately     */
#define KERN_CRIT       "<2>"   /* critical conditions                  */
#define KERN_ERR        "<3>"   /* error conditions                     */
#define KERN_WARNING    "<4>"   /* warning conditions                   */
#define KERN_NOTICE     "<5>"   /* normal but significant condition     */
#define KERN_INFO       "<6>"   /* informational                        */
#define KERN_DEBUG      "<7>"   /* debug-level messages                 */

#define printk printf



/*  VFS to GFS abstraction  */

/*  Convert a "struct mount" in to a GFS superblock.  */
#define vsb2sbd(vsbp) ((gfs_sbd_t *)((vsbp)->mnt_data)

/*  Convert a "vnode" into a GFS inode.  */
#define vi2ino(vip) ((gfs_inode_t *)((vip)->v_data))
#define vtogi(vp) vi2ino(vp)

/*  Convert a linux Buffer head into GFS buffer head data */
#define bh2pd(bh) ((gfs_bhdata_t *)((bh)->b_driver1))



/*  GFS to VFS abstraction  */

typedef struct mount gfs_vsb_t;
typedef struct vnode gfs_vino_t;



/*  Buffer head abstraction 
    
    All macros take a pointer to a buffer_head_t structure.

    BH_DATA() returns a pointer to the data held in the buffer.
    BH_SIZE() returns the size in bytes of the data in the buffer.
    BH_COUNT() is the usage count of the buffer.
    */

typedef struct buf buffer_head_t;
#define BH_DATA(x) ((x)->b_data)
#define BH_SIZE(x) ((x)->b_bcount)
#define BH_BLKNO(x) ((x)->b_blkno)



/*  Endianness conversion abstraction  */

#define __BYTE_ORDER __LITTLE_ENDIAN

#include "gfs_endian.h"



/*  Directory translation abstraction  */

typedef int (*filldir_t)(void *, const char *, int, off_t, ino_t);



/*  struct qstr definition 
 
    "quick string" -- eases parameter passing, but more importantly
    saves "metadata" about the string (ie length and the hash).
    */

struct qstr {
        const unsigned char * name;
        unsigned int len;
        unsigned int hash;
};



/*  Semaphore abstraction

    All macros take a pointer to a semaphore_t structure.
    GFS_SEMA_DOWN_TRY() returns FALSE if the down() succeeds 
    */

typedef int semaphore_t;
#define GFS_SEMA_DOWN(x) {IFDEBUG(DEBUG_SEMAPHORE, \
				  printf("(%u) Down: <%p> %u\t%s\t\"%s\"\n",(uint)kpid(),(void *) x, __LINE__, __FILE__, #x);); \
			  IFDEBUG(DEBUG_SEMAPHORE, \
				  printf("(%u) done\n", (uint)kpid());); \
			 }
#define GFS_SEMA_DOWN_INTERRUPTIBLE(x) {IFDEBUG(DEBUG_SEMAPHORE, \
				  printf("(%u) Down: <%p> %u\t%s\t\"%s\"\n",(uint)kpid(),(void *) x, __LINE__, __FILE__, #x);); \
			  IFDEBUG(DEBUG_SEMAPHORE, \
				  printf("(%u) done\n", (uint)kpid());); \
			 }
#define GFS_SEMA_DOWN_TRY(x) 0
#define GFS_SEMA_UP(x) {IFDEBUG(DEBUG_SEMAPHORE, \
				printf("(%u) Up:   <%p> %u\t%s\t\"%s\"\n",(uint)kpid(),(void *)x, __LINE__, __FILE__, #x);); \
			IFDEBUG(DEBUG_SEMAPHORE, \
				printf("(%u) done\n", (uint)kpid());); \
		       }

#define GFS_SEMA_INIT(x)
#define GFS_SEMA_INIT_LKD(x)



/*  PID abstraction

    Returns the PID of the current process.
    */

#define kpid() curproc->p_pid



/*  CURRENT_TIME macro 
    Returns the standard UNIX time integer 

    CURRENT_TIME is already defined in Linux.
    */

#define CURRENT_TIME time_second



/*  FSBTOBB() definition  */

#define FSBTOBB(b)  ((b) << sdp->sd_fsb2bb_shift)



/*  Other Stuff  */

#define KRET(x) x

#define semaphore sem

typedef off_t loff_t;


#endif
