/* #include "opt_debug_gfs.h" */

#define EXTERN

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <miscfs/specfs/specdev.h> /* XXX */	/* defines v_rdev */
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/fcntl.h>
#include <sys/conf.h>
/* #include <miscfs/nullfs/null.h> */

#include <lgfs.h>
#include <gfs_fbsd.h>
#include <incore.h>
#include <gfs_debug.h>
#include <gfs.h>
/* #include <inode.h> */
#include <incore.h>
#include <blklist.h>
#include <fs.h>
#include <locking.h>
#include <glock.h>
#include <dio.h>


#include <super.h>

#define GFS_DIAGNOSTIC

static MALLOC_DEFINE(M_GFSMNT, "GFS mount", "GFS mount structure");

static int	gfs_fhtovp __P((struct mount *mp, struct fid *fidp,
							struct sockaddr *nam, struct vnode **vpp,
							int *exflagsp, struct ucred **credanonp));
static int	gfs_mount __P((struct mount *mp, char *path, caddr_t data,
						   struct nameidata *ndp, struct proc *p));
static int	gfs_quotactl __P((struct mount *mp, int cmd, uid_t uid,
							  caddr_t arg, struct proc *p));
static int	gfs_root __P((struct mount *mp, struct vnode **vpp));
static int	gfs_start __P((struct mount *mp, int flags, struct proc *p));
static int	gfs_statfs __P((struct mount *mp, struct statfs *sbp,
							struct proc *p));
static int	gfs_sync __P((struct mount *mp, int waitfor,
						  struct ucred *cred, struct proc *p));
static int	gfs_unmount __P((struct mount *mp, int mntflags,
							 struct proc *p));
static int	gfs_vget __P((struct mount *mp, ino_t ino,
						  struct vnode **vpp));
static int	gfs_vptofh __P((struct vnode *vp, struct fid *fhp));


/* super.c */
void gfs_clear_out_gla(gfs_sbd_t *sdp);
#if 0
int  gfs_mountbit_find(gfs_sbd_t *sdp);
int  gfs_mountbit_value(uint8 *bitmap, unsigned int no);
void gfs_mountbit_set(uint8 *bitmap, unsigned int no, int value);
#endif


static int
gfs_mount(struct mount *mp,
		  char *path,
		  caddr_t data,
		  struct nameidata *ndp,
		  struct proc *p){
  size_t size;
  int x,error = 0;
  gfs_args_t args;
  struct vnode    *devvp;

  gfs_sbd_t *sdp;
  gfs_dcached_t *dcd;

  int silent = FALSE;
  /* */
  int ronly, ncount;

  IFDEBUG(DEBUG_SUPER1,
		  printf(DEBUG_HOW "Entering gfs_mount()\n"););

#ifdef GFS_DIAGNOSTIC
  printf("gfs_mount(mp = %p)\n", (void *)mp);
  printf("gfs_mount(path = %s)\n", path);
#endif

  /*
   * Get argument
   */
  error = copyin(data, (caddr_t)&args, sizeof(gfs_args_t));
  if (error)
	return (error);
  printf ("gfs_mount args :%s\n%s\n%s\n%s\n",args.target,
		  args.lockproto,args.locktable,args.hostdata);

  if (mp->mnt_flag & MNT_UPDATE) {
	printf("gfs_mount: MNT_UPDATE not implemented yet.\n");
	return (EINVAL);
  }

  /*  NDINIT(ndp, LOOKUP, FOLLOW|WANTPARENT|LOCKLEAF, UIO_USERSPACE, args.target, p); */
  NDINIT(ndp, LOOKUP, FOLLOW, UIO_USERSPACE, args.target, p); 
  error = namei(ndp);
  if(error){
	printf("Error doing lookupname: %d\n", error);
	return (error);
  }

  devvp = ndp->ni_vp;

  if (devvp->v_type != VBLK) {
	printf("GFS_mount: error not a block device %d\n",error);
	return(ENOTBLK);
  }
#if 0
  if (major(devvp->v_rdev) >= nblkdev ||
	  bdevsw(devvp->v_rdev) == NULL) {
	vrele(devvp);
	return (ENXIO);
  }
#endif
  if (bdevsw(devvp->v_rdev) == NULL) {
	vrele(devvp);
	return (ENXIO);
  }
  /* taken from ntfs */
  (void) copyinstr(path, 
				   mp->mnt_stat.f_mntonname,
				   MNAMELEN - 1,
				   &size);
  bzero(mp->mnt_stat.f_mntonname + size, MNAMELEN - size);

  (void) copyinstr(args.target,
				   mp->mnt_stat.f_mntfromname, MNAMELEN - 1,
				   &size);
  bzero(mp->mnt_stat.f_mntfromname + size, MNAMELEN - size);


  /* taken from ntfs */
  error = vfs_mountedon(devvp);
  if (error)	return (error);

  ncount = vcount(devvp);

  if (devvp->v_object)	ncount -= 1;
  if (ncount > 1 && devvp != rootvp) 	return (EBUSY);
#ifndef DEBUG_LOCKS
  vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY, p);
#else
  debug_vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY, p,__FILE__,__LINE__);
#endif
  error = vinvalbuf(devvp, V_SAVE, p->p_ucred, p, 0, 0);
  VOP_UNLOCK(devvp, 0, p);
  if (error)	return (error);
  
  ronly = (mp->mnt_flag & MNT_RDONLY) != 0;
  error = VOP_OPEN(devvp, ronly ? FREAD : FREAD|FWRITE, FSCRED, p);

  if (error)	return (error);
  /*  taken from ntfs */

  sdp = (gfs_sbd_t *)gmalloc(sizeof(gfs_sbd_t));
  if (!sdp){
	printf("GFS: out of memory\n");
	goto fail;
  }
  bzero(sdp,sizeof(gfs_sbd_t));

  sdp->sd_dev = devvp;

#if defined(__linux__)
  vsb2sbd(sb) = sdp;
  sdp->sd_vsb = sb;
#endif
#if defined(__FreeBSD__)
  mp->mnt_data = (qaddr_t)sdp;
  sdp->sd_vsb = mp;
#endif

  /*  Init semaphores  */

  for (x = 0; x < GFS_GHASH_SIZE; x++)
	{
	  INIT_GFS_LIST_HEAD(&sdp->sd_gla[x]);
	  GFS_SEMA_INIT(&sdp->sd_gla_lock[x]);
	}
  INIT_GFS_LIST_HEAD(&sdp->sd_gl_lru);
  GFS_SEMA_INIT(&sdp->sd_gl_lru_lock);
  
  for (x = 0; x < GFS_IHASH_SIZE; x++)
	{
	  INIT_GFS_LIST_HEAD(&sdp->sd_ihash[x]);
	  GFS_SEMA_INIT(&sdp->sd_ihash_lock[x]);
	}
  
  if (gfs_mount_lockproto(sdp, &args))
    goto fail_sbd_free;
  
  error = gfs_glock(sdp, GFS_SB_LOCK, NULL, 0);
  if (error < 0)	{
	  printf("GFS:  can't get root glock\n");
	  goto fail_lock_unmount;
  }
  
  if (sget(sdp, silent) < 0){
	printf("GFS: sget failed\n");
	error = EINVAL;
    goto fail_gla_free;
  }

  /*  Set mount information  */
  if (gfs_mountbit_find(sdp) < 0)  {
    printf("GFS:  too many mounts on this filesystem\n");
    goto fail_ri_free;
  }

#if defined(__linux__)
  sdp->sd_fsb2bb_shift = 0;
#endif  /*  linux  */
#if defined(__sgi__) || defined(__FreeBSD__)
  sdp->sd_fsb2bb_shift = ffs(sdp->sd_sb.sb_bsize / GFS_BASIC_BLOCK) - 1;
#endif  /*  sgi  */
  sdp->sd_diptrs = (sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t)) / sizeof(uint64);
  sdp->sd_inptrs = sdp->sd_sb.sb_bsize / sizeof(uint64);

  if (gfs_init_tunevals(sdp, 1))
	goto fail_ri_free;

  dcd = (gfs_dcached_t *)gmalloc(sizeof(gfs_dcached_t));
  if (!dcd)
    goto fail_dput;
  
  /*  dcd->dc_vn = vi2ino(sdp->sd_rooti)->i_vn; */
  dcd->dc_time = CURRENT_TIME;
  
  mp->mnt_stat.f_fsid.val[0] = dev2udev(devvp->v_rdev);
  mp->mnt_stat.f_fsid.val[1] = mp->mnt_vfc->vfc_typenum;
  mp->mnt_maxsymlinklen = 0;
  mp->mnt_flag |= MNT_LOCAL;
  devvp->v_specmountpoint = mp;


  if (sput(sdp) < 0)
    goto fail_inoded;
  
  gfs_gunlock(sdp, GFS_SB_LOCK, 0);

  /*   (void)gfs_statfs(mp, &mp->mnt_stat, p); */
  printf("Exiting gfs_mount() (ok)\n");
  return(0); 
  
 fail_inoded:
  sdp->sd_id_stop = 1;
  GFS_SEMA_UP(&sdp->sd_id_work);
  GFS_SEMA_DOWN(&sdp->sd_id_lock);

 fail_bufd:
  sdp->sd_bufd_stop = 1;
  GFS_SEMA_DOWN(&sdp->sd_bufd_lock);

 fail_dput:
  /*  dput(sb->s_root); */
  goto fail_ri_free;

 fail_iput:
  /*  iput(sdp->sd_rooti); */

 fail_tune_free:
  gfree(sdp->tune, sizeof(gfs_tune_t));

 fail_ri_free:  
  gfree(sdp->sd_rgd,0); /* size not used by bsd */

 fail_gla_free:
  gfs_gunlock(sdp, GFS_SB_LOCK, 0);
  gfs_clear_out_gla(sdp); 

 fail_lock_unmount:
  gfs_unmount_lockproto(sdp);
  
 fail_sbd_free:
  gfree(sdp,0); /* size not used by bsd */
  
 fail:
  devvp->v_specmountpoint = NULL;
  /*   if (bp)	brelse(bp); */
  (void)VOP_CLOSE(devvp, ronly ? FREAD : FREAD|FWRITE, NOCRED, p);
  printf("Exiting gfs_mount() (error)\n");

  return (error);
}

static int
gfs_start(mp, flags, p)
	 struct mount *mp;
	 int flags;
	 struct proc *p;
{
  printf("gfs_start\n");
  return (0);
  /* return VFS_START(MOUNTTONULLMOUNT(mp)->gfsm_vfs, flags, p); */
}

static int
gfs_unmount(mp, mntflags, p)
	 struct mount *mp;
	 int mntflags;
	 struct proc *p;
{
  struct vnode *gfsm_rootvp = MOUNTTOGFSMOUNT(mp)->gfsm_rootvp;
  /*  int error; unused? */
  int flags = 0;

#ifdef GFS_DIAGNOSTIC
  printf("gfs_unmount(mp = %p)\n", (void *)mp);
#endif

  if (mntflags & MNT_FORCE)
	flags |= FORCECLOSE;

  /*
   * Release reference on underlying root vnode
   */
  vrele(gfsm_rootvp);
  /*
   * And blow it away for future re-use
   */
  vgone(gfsm_rootvp);
  /*
   * Finally, throw away the gfs_mount structure
   */
  free(mp->mnt_data, M_GFSMNT);	/* XXX */
  mp->mnt_data = 0;
  return 0;
}

static int
gfs_root(mp, vpp)
	 struct mount *mp;
	 struct vnode **vpp;
{
  struct vnode *nvp;
  int error;

  gfs_sbd_t *sdp = (gfs_sbd_t *)mp->mnt_data;

	/*  struct proc *p = curproc; *//* unused? */
  /*  struct vnode *vp; unused? */
  IFDEBUG(DEBUG_SUPER1,
		  printf(DEBUG_HOW "Entering gfs_root()\n"););

  printf("gfs_root(mp = %p, gfs_sbd_t = %p )\t", (void *)mp,(void *)sdp);
  printf("inode sdp->sd_sb.sb_rootdi %u\n",(ino_t)sdp->sd_sb.sb_rootdi);

  /* note sb_rootdi is a 64 bit number ino_t is 32... shouldn't cause problem here */ 
  error = VFS_VGET(mp, (ino_t)sdp->sd_sb.sb_rootdi, &nvp);

  if (error)
	return (error);
  *vpp = nvp;
  /*  *vpp = sdp->sd_dev; */
  
  IFDEBUG(DEBUG_SUPER1,
		  printf(DEBUG_HOW "Exiting gfs_root()\n"););

  return 0;
}

static int
gfs_quotactl(mp, cmd, uid, arg, p)
	 struct mount *mp;
	 int cmd;
	 uid_t uid;
	 caddr_t arg;
	 struct proc *p;
{
  return (0);
}

static int
gfs_statfs(struct mount *mp,
		   struct statfs *sfs,
		   struct proc *p){
  int error;
  /*  struct statfs *mstat; */

  gfs_sbd_t *sdp;
  gfs_rgrpd_t *rgd;
  /*  struct statfs sfs; */
  int i;
  uint64 blocks = 0;
  uint64 dinodes = 0;
  uint64 free = 0;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_statfs()\n"););

#if 0
  bzero(&mstat, sizeof(mstat));
  mstat = MOUNTTOGFSMOUNT(mp);
#endif
  /* now copy across the "interesting" information and fake the rest */
  if (sfs != &mp->mnt_stat) {
	bcopy(&mp->mnt_stat.f_fsid,       &sfs->f_fsid,        sizeof(sfs->f_fsid));
	bcopy( mp->mnt_stat.f_mntonname,   sfs->f_mntonname,   MNAMELEN);
	bcopy( mp->mnt_stat.f_mntfromname, sfs->f_mntfromname, MNAMELEN);
  }
  
  if (!mp || !(sdp = (gfs_sbd_t *)mp->mnt_data))
    return KRET(EIO);

  /*   memset(&sfs, 0, sizeof(struct statfs)); */

  for (i = 0; i < sdp->sd_sb.sb_rgrps; i++){
    rgd = &sdp->sd_rgd[i];
    error = gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, GL_SHARED);
    if (error){
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting gfs_statfs()\n"););
      return(error);      
    }
    /*  Read/Reread the RG block  */
    error = gfs_rgrp_read(sdp, rgd);
    if (error) {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting gfs_statfs()\n"););
      return(error);
    }
      
    IFDEBUG(DEBUG_SUPER2,
			printk(DEBUG_HOW "%.4u:  %u%u\t%u%u\t%u%u\n",
				   i,
				   (uint)rgd->rd_rg.rg_dinodes >> 32,(uint)rgd->rd_rg.rg_dinodes,
				   (uint)(rgd->rd_ri.ri_data - rgd->rd_rg.rg_free) >> 32 ,(uint)(rgd->rd_ri.ri_data - rgd->rd_rg.rg_free),
				   (uint)rgd->rd_ri.ri_data >> 32 ,(uint)rgd->rd_ri.ri_data););

    blocks += rgd->rd_ri.ri_data;
    dinodes += rgd->rd_rg.rg_dinodes;
    free += rgd->rd_rg.rg_free;
	
    gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
  }

  /* Do we need to set these?
  sbp->f_flags = mstat->f_flags;
  sbp->f_iosize = mstat->f_iosize;
  */
  printf ("sfs->f_flags %d\n",sfs->f_flags);
  printf ("mp->mnt_flag %d\n",mp->mnt_flag);
  sfs->f_flags = mp->mnt_flag;
  sfs->f_type = GFS_SUPER_MAGIC;
  sfs->f_bsize = sdp->sd_sb.sb_bsize;
  sfs->f_blocks = blocks;
  sfs->f_bfree = free;
  sfs->f_bavail = free;
  sfs->f_files = dinodes;
  sfs->f_ffree = free;
  /*  sfs->f_namelen = GFS_FNAMESIZE; */

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_statfs()\n"););

  /*  return(copy_to_user(statfsbuf, &sfs, (size < sizeof(struct statfs)) ? size : sizeof(struct statfs))); */
  return(0);
}

static int
gfs_sync(mp, waitfor, cred, p)
	 struct mount *mp;
	 int waitfor;
	 struct ucred *cred;
	 struct proc *p;
{
  /*
   * XXX - Assumes no data cached at null layer.
	 */
  return (0);
}

static int
gfs_vget(struct mount *mp,
		 ino_t ino,
		 struct vnode **vpp){

  gfs_inode_t *ip, *ret_ip;
  gfs_sbd_t *sdp = (gfs_sbd_t *)mp->mnt_data;
  uint64 glock;
  int error;

  struct vnode *vp;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_vget()\n"););
  ip = (gfs_inode_t *)gmalloc(sizeof(gfs_inode_t));
  if (!ip) {
    printk(KERN_ERR "GFS:  out of memory\n");
    goto fail;
  }
  
  memset(ip, 0, sizeof(gfs_inode_t));

  
  /* vi2ino(inode) = ip; */
  /*  ip->i_vino = inode; */

  ip->i_sbd = sdp;
  ip->i_num = ino;
  /*
  ip->i_num = inode->i_ino;
  ip->i_sema = MUTEX;
  ip->i_diroff = NULL;
  */

  /*  Aquire glock  */

  glock = gfs_ino2gl(sdp, ip->i_num);

  if ((error = gfs_glock(sdp, glock, NULL, 0)) < 0)
    goto fail_ino_free;

  if ((ret_ip = gfs_ihashlookup(sdp, ip->i_num)) != NULL){
	printf("gfs_vget... inode in hash already\n");
	VREF(ret_ip->i_vino);
	*vpp = ret_ip->i_vino;
	/* 	Debugger("gfs_vget"); */ /* why are we back here? */
	/* need some check for unlock here... lookup more at irix code */
	gfs_gunlock(sdp, glock, 0);
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_vget() (OK)\n"););
	return (0);
  }
  /*  ASSERT(gfs_ihashlookup(sdp, ip->i_num) == NULL); */

  error = getnewvnode(VT_NON, mp, gfs_vnodeop_p, &vp);
  /*  printf("gfs_vget new vnode (%p) error %u\n",vp,error); */
  vprint("gfs_vget new vnode\n",vp);
  if (error){
#if 0
	if (ffs_inode_hash_lock < 0)
	  wakeup(&ffs_inode_hash_lock);
	ffs_inode_hash_lock = 0;
#endif
	*vpp = NULL;
	gfree(ip,0); /* size not used by bsd */
	return (error);
  }
  ip->i_vino = vp;
  vp->v_data = ip;

  /*  Read in dinode  */
  /*  Debugger("gfs_vget"); */
  ip->i_didata = gfs_dread(sdp, ip->i_num, glock);
  if (!ip->i_didata){
	vput(vp); 
    goto fail_gunlock;
  }

  gfs_dinode_in(&ip->i_di, ip->i_didata->b_data);

  if (ip->i_di.di_magic != GFS_MAGIC_DI){
    printk("GFS:  ip->i_di.di_magic (%"FMT_64"u) != GFS_MAGIC_DI\n",ip->i_num);
    {
      uint32 x;
      for (x = 0; x < sdp->sd_sb.sb_bsize; x++) {
		printk("%.2X ", ((unsigned char *)(ip->i_didata->b_data))[x]);
		if (x % 16 == 15)
		  printk("\n");
      }
      if (x % 16 != 0)
		printk("\n");
    }
    ASSERT(0);
  }
  ASSERT(ip->i_di.di_addr == ip->i_num);
  
  IFDEBUG(DEBUG_SUPER3,
		  int x;
		  /*		  printk(DEBUG_HOW "gfs_read_inode(%"FMT_64"u)\n", ip->i_num); */
		  printk(DEBUG_HOW "gfs_vget(%u)\n", (uint)ip->i_num); 
		  printk(DEBUG_HOW "Before:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
		  printk(DEBUG_HOW "Before:  ip->i_di.di_open = ");
		  for (x = 0; x < GFS_MOUNT_BITS; x++)
		  printk("%.2x ", ip->i_di.di_open[x]);
		  printk("\n");
		  printk(DEBUG_HOW "Before:  sdp->sd_id = %u\n",sdp->sd_id););
  
  ip->i_di.di_nopen++;
  /* VNODE */
  vp->v_type = VNON; /* just set it to nothing for the moment */
  vp->v_type = VDIR; /* OK force it to somethign it likes */
  /* vp->v_usecount =  ip->i_di.di_nopen++; */
  if(ino == (ino_t)sdp->sd_sb.sb_rootdi){
	vp->v_flag |= VROOT;
  }

  
  /*  ASSERT(!gfs_mountbit_value(ip->i_di.di_open, sdp->sd_id)); */
  gfs_mountbit_set(ip->i_di.di_open, sdp->sd_id, 1);
  
  IFDEBUG(DEBUG_SUPER3,
		  int x;
		  printk(DEBUG_HOW "After:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
		  printk(DEBUG_HOW "After:  ip->i_di.di_open = ");
		  for (x = 0; x < GFS_MOUNT_BITS; x++)
		  printk("%.2x ", ip->i_di.di_open[x]);
		  printk("\n");
		  printk(DEBUG_HOW "After:  sdp->sd_id = %u\n",sdp->sd_id););
    
  /*  Write dinode back out  */
  
  gfs_dinode_out(&ip->i_di, ip->i_didata->b_data);
#if 1 /* not for initial testing */
  gfs_dwrite(ip->i_didata, TRUE); 
#endif
  gfs_inode_attr_in(ip);
  
  gfs_ihashadd(ip);
  
  gfs_gunlock(sdp, glock, 0);
  
#if 0
  if (S_ISREG(inode->i_mode))
    inode->i_op = &gfs_file_iops;
  else if (S_ISDIR(inode->i_mode))
    inode->i_op = &gfs_dir_iops;
  else if (S_ISLNK(inode->i_mode))
    inode->i_op = &gfs_symlink_iops;
  else
    panic("GFS:  unknown file type:  %d\n", inode->i_mode);
#endif

  *vpp = vp; 
   IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_vget() (OK)\n"););
  return(0);
  
 fail_gunlock:
  gfs_gunlock(sdp, glock, 0);
  
 fail_ino_free:
  gfree(ip,0); /*size not used by bsd */
  
 fail:
#if 0
  make_bad_inode(inode);
#endif

  printk("GFS:  making bad inode\n");
  
  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting  gfs_vget() (ERROR)\n"););
  return(0);
}

static int
gfs_fhtovp(mp, fidp, nam, vpp, exflagsp, credanonp)
	 struct mount *mp;
	 struct fid *fidp;
	 struct sockaddr *nam;
	 struct vnode **vpp;
	 int *exflagsp;
	 struct ucred**credanonp;
{
  
  printf("gfs_fhtovp\n");
  return (0);
}

static int
gfs_vptofh(vp, fhp)
	 struct vnode *vp;
	 struct fid *fhp;
{
  printf ("gfs_vptofh\n");
  return (0);
}

static struct vfsops gfs_vfsops = {
  gfs_mount,
  gfs_start,
  gfs_unmount,
  gfs_root,
  gfs_quotactl,
  gfs_statfs,
  gfs_sync,
  gfs_vget,
  gfs_fhtovp,
  gfs_vptofh,
  gfs_init,
};

VFS_SET(gfs_vfsops, gfs, VFCF_LOOPBACK);


/*
 * Initialise cache headers
 */
int
gfs_init(struct vfsconf *vfsp){
  int error;
  GFS_dflags = 0xFFFFFFFFFFFFFFFFll & ~( DEBUG_SEMAPHORE | 
										 /* DEBUG_GLOCK1 | */ 
										 DEBUG_ONDISK1 |
										 DEBUG_SUPER2 );
  printf("gfs_init\n");		/* printed during system boot */
  Debugger("gfs_init");  /* allow us put into gdb mode */
  if(error = nolock_init_module()){
	printf("nolock_init_module failed %d\n",error);
	return(error);
  } else {
	printf("Installed the nolock module\n");
  }
  return (0);
}
