/* #include "opt_debug_gfs.h" */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/unistd.h>
#include <sys/sysctl.h>
#include <sys/vnode.h>
#include <sys/dirent.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/malloc.h>
#include <sys/buf.h>

#include <machine/param.h>
/* #include <miscfs/nullfs/null.h> */
#include <lgfs.h>
#include <gfs_fbsd.h>
#include <incore.h>
#include <gfs_debug.h>
#include <gfs.h>
#include <inode.h>
#include <incore.h>
#include <blklist.h>
#include <fs.h>
#include <dir.h>
#include <glock.h>
#include <readdir.h>


int	prtactive = 1;		/* 1 => print out reclaim of active vnodes */

static int gfs_bug_bypass = 0;   /* for debugging: enables bypass printf'ing */
SYSCTL_INT(_debug, OID_AUTO, gfs_bug_bypass, CTLFLAG_RW, &gfs_bug_bypass, 0, "");

static int	gfs_access __P((struct vop_access_args *ap));
static int	gfs_bwrite __P((struct vop_bwrite_args *ap));
static int	gfs_getattr __P((struct vop_getattr_args *ap));
static int	gfs_inactive __P((struct vop_inactive_args *ap));
static int	gfs_lock __P((struct vop_lock_args *ap));
static int	gfs_lookup __P((struct vop_lookup_args *ap));
static int	gfs_print __P((struct vop_print_args *ap));
static int	gfs_reclaim __P((struct vop_reclaim_args *ap));
static int	gfs_setattr __P((struct vop_setattr_args *ap));
static int	gfs_strategy __P((struct vop_strategy_args *ap));
static int	gfs_unlock __P((struct vop_unlock_args *ap));

static int gfs_missingop __P((struct vop_generic_args *ap));
static int gfs_create __P((struct vop_create_args *));
static int gfs_abortop __P((struct vop_abortop_args *));
static int gfs_advlock __P((struct vop_advlock_args *));
       int gfs_bmap __P((struct vop_bmap_args *));
static int gfs_close __P((struct vop_close_args *));
static int gfs_link __P((struct vop_link_args *));
static int gfs_mkdir __P((struct vop_mkdir_args *));
static int gfs_mknod __P((struct vop_mknod_args *));
static int gfs_mmap __P((struct vop_mmap_args *));
static int gfs_open __P((struct vop_open_args *));
static int gfs_pathconf __P((struct vop_pathconf_args *));
static int gfs_readdir __P((struct vop_readdir_args *));
static int gfs_readlink __P((struct vop_readlink_args *));
static int gfs_remove __P((struct vop_remove_args *));
static int gfs_rename __P((struct vop_rename_args *));
static int gfs_rmdir __P((struct vop_rmdir_args *));
static int gfs_symlink __P((struct vop_symlink_args *));
static int gfs_whiteout __P((struct vop_whiteout_args *));


int
gfs_bypass(ap)
	 struct vop_generic_args /* {
								struct vnodeop_desc *a_desc;
								<other random data follows, presumably>
								} */ *ap;
{
  printf ("gfs_bypass: \n");
	
  return (0);
}

static int
gfs_missingop(ap)
	struct vop_generic_args *ap;
{
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_missingop()\n"););
  printf("no vop function for %s in gfs child", ap->a_desc->vdesc_name);
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_missingop(OK)\n"););
  return (EOPNOTSUPP);
}

/*
 * Bmap converts a the logical block number of a file to its physical block
 * number on the disk. The conversion is done by using the logical block
 * number to index into the array of block pointers described by the dinode.
 */
int
gfs_bmap(ap)
	struct vop_bmap_args /* {
		struct vnode *a_vp;
		ufs_daddr_t a_bn;
		struct vnode **a_vpp;
		ufs_daddr_t *a_bnp;
		int *a_runp;
		int *a_runb;
	} */ *ap;
{
	/*
	 * Check for underlying vnode requests and ensure that logical
	 * to physical mapping is requested.
	 */
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_bmap()\n"););
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_bmap(OK)\n"););
	return (EINVAL);
}

/*
 * Mknod vnode call
 */
/* ARGSUSED */
int
gfs_mknod(ap)
	struct vop_mknod_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap;
{
#if 0
	struct vattr *vap = ap->a_vap;
	struct vnode **vpp = ap->a_vpp;
	struct inode *ip;
#endif
	int error = EINVAL;

	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_mknod()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_mknod(OK)\n"););
	return (error);
}
/*
 * Mmap a file
 *
 * NB Currently unsupported.
 */
/* ARGSUSED */
int
gfs_mmap(ap)
	struct vop_mmap_args /* {
		struct vnode *a_vp;
		int  a_fflags;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_mmap()\n"););
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_mmap(OK)\n"););
	return (EINVAL);
}

int
gfs_remove(ap)
	struct vop_remove_args /* {
		struct vnode *a_dvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap;
{
#if 0
	struct inode *ip;
	struct vnode *vp = ap->a_vp;
	struct vnode *dvp = ap->a_dvp;
#endif
	int error = EINVAL;
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_remove()\n"););
	IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_remove(OK)\n"););
	return (error);
}

int
gfs_rename(ap)
	struct vop_rename_args  /* {
		struct vnode *a_fdvp;
		struct vnode *a_fvp;
		struct componentname *a_fcnp;
		struct vnode *a_tdvp;
		struct vnode *a_tvp;
		struct componentname *a_tcnp;
	} */ *ap;
{
#if 0
	struct vnode *tvp = ap->a_tvp;
	register struct vnode *tdvp = ap->a_tdvp;
	struct vnode *fvp = ap->a_fvp;
	struct vnode *fdvp = ap->a_fdvp;
	struct componentname *tcnp = ap->a_tcnp;
	struct componentname *fcnp = ap->a_fcnp;
	struct proc *p = fcnp->cn_proc;
#endif
	int error = EINVAL;
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_rename()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_rename(OK)\n"););
	return(error);
}

/*
 * Open called.
 *
 * Nothing to do.
 */
/* ARGSUSED */
int
gfs_open(ap)
	struct vop_open_args /* {
		struct vnode *a_vp;
		int  a_mode;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_open()\n"););
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_open(OK)\n"););
	return (0);
}

/*
 * Close called.
 *
 * Update the times on the inode.
 */
/* ARGSUSED */
int
gfs_close(ap)
	struct vop_close_args /* {
		struct vnode *a_vp;
		int  a_fflag;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap;
{
#if 0
  register struct vnode *vp = ap->a_vp;
#endif
  
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_close()\n"););
  simple_lock(&vp->v_interlock);
#if 0
  if (vp->v_usecount > 1)
	gfs_itimes(vp);
#endif
  simple_unlock(&vp->v_interlock);
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_close(OK)\n"););
  return (0);
}

/*
 * Return POSIX pathconf information applicable to ufs filesystems.
 */
int
gfs_pathconf(ap)
	struct vop_pathconf_args /* {
		struct vnode *a_vp;
		int a_name;
		int *a_retval;
	} */ *ap;
{
  int error = 0;
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_pathconf()\n"););

  switch (ap->a_name) {
	case _PC_LINK_MAX:
		*ap->a_retval = LINK_MAX;
		break;	
	case _PC_NAME_MAX:
		*ap->a_retval = NAME_MAX;
		break;
	case _PC_PATH_MAX:
		*ap->a_retval = PATH_MAX;
		break;	
	case _PC_PIPE_BUF:
		*ap->a_retval = PIPE_BUF;
		break;
	case _PC_CHOWN_RESTRICTED:
		*ap->a_retval = 1;
		break;
	case _PC_NO_TRUNC:
		*ap->a_retval = 1;
		break;
	default:
		error = EINVAL;
		break;
	}
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_pathconf(OK)\n"););
	return(error);
}


/*
 * Ufs abort op, called after namei() when a CREATE/DELETE isn't actually
 * done. If a buffer has been saved in anticipation of a CREATE, delete it.
 */
/* ARGSUSED */
int
gfs_abortop(ap)
	struct vop_abortop_args /* {
		struct vnode *a_dvp;
		struct componentname *a_cnp;
	} */ *ap;
{
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_abortop)\n"););
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_abortop(OK)\n"););
	return (0);
}
/*
 * Advisory record locking support
 */
int
gfs_advlock(ap)
	struct vop_advlock_args /* {
		struct vnode *a_vp;
		caddr_t  a_id;
		int  a_op;
		struct flock *a_fl;
		int  a_flags;
	} */ *ap;
{
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_advlock()\n"););
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_advlock(OK)\n"););

	return (EINVAL);
}
/*
 * link vnode call
 */
int
gfs_link(ap)
	struct vop_link_args /* {
		struct vnode *a_tdvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap;
{
#if 0
	struct vnode *vp = ap->a_vp;
	struct vnode *tdvp = ap->a_tdvp;
	struct componentname *cnp = ap->a_cnp;
	struct proc *p = cnp->cn_proc;
#endif
	int error = EINVAL;

	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_link()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_link(OK)\n"););
	return (error);
}

/*
 * whiteout vnode call
 */
int
gfs_whiteout(ap)
	struct vop_whiteout_args /* {
		struct vnode *a_dvp;
		struct componentname *a_cnp;
		int a_flags;
	} */ *ap;
{
#if 0
	struct vnode *dvp = ap->a_dvp;
	struct componentname *cnp = ap->a_cnp;
#endif
	int error = 0;
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_whiteout()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_whiteout(OK)\n"););
	return (error);
}
/*
 * Mkdir system call
 */
int
gfs_mkdir(ap)
	struct vop_mkdir_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap;
{
#if 0
	register struct vnode *dvp = ap->a_dvp;
	register struct vattr *vap = ap->a_vap;
	register struct componentname *cnp = ap->a_cnp;
	register struct inode *ip, *dp;
	struct vnode *tvp;
	int dmode;
#endif
	int error = EINVAL;

	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_mkdir()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_mkdir(OK)\n"););
	return (error);
}

/*
 * Rmdir system call.
 */
int
gfs_rmdir(ap)
	struct vop_rmdir_args /* {
		struct vnode *a_dvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap;
{
#if 0
	struct vnode *vp = ap->a_vp;
	struct vnode *dvp = ap->a_dvp;
	struct componentname *cnp = ap->a_cnp;
	struct inode *ip, *dp;
#endif
	int error = EINVAL;
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_rmdir()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_rmdir(OK)\n"););
	return (error);
}

/*
 * symlink -- make a symbolic link
 */
int
gfs_symlink(ap)
	struct vop_symlink_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
		char *a_target;
	} */ *ap;
{
#if 0
	register struct vnode *vp, **vpp = ap->a_vpp;
	register struct inode *ip;
	int len;
#endif
	int error = 0;
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_symlink()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_sysmlink(OK)\n"););
	return (error);
}


/*
 * Vnode op for reading directories.
 *
 * The routine below assumes that the on-disk format of a directory
 * is the same as that defined by <sys/dirent.h>. If the on-disk
 * format changes, then it will be necessary to do a conversion
 * from the on-disk format that read returns to the format defined
 * by <sys/dirent.h>.
 */
int
gfs_readdir(ap)
	struct vop_readdir_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
		int *a_eofflag;
		int *ncookies;
		u_long **a_cookies;
	} */ *ap;
{
	struct vnode *vp = ap->a_vp;
	gfs_inode_t *dip = vtogi(vp);
	int error = 0;
	gfs_sbd_t *sdp = dip->i_sbd;

	int64 f_pos = 0; 
	void *dirent;
	void *_buf;
	struct dirent dp;
	register struct uio *uio = ap->a_uio;
	int	startpos,pos;
	int reclen;
	off_t off, offset;
	
	struct fakedirent fent;

	fent.f_uio = uio;
	fent.f_dirent = &dp;

	pos = 0;
 
#if 0
	size_t count, lost;
#endif

	startpos = uio->uio_offset;

	IFDEBUG(DEBUG_FILE1,
			printk(DEBUG_HOW "Entering gfs_readdir()\n");); 
	printf("gfs_readdir uio->uio_offset %u uio->uio_resid %u\n",(u_int)uio->uio_offset,uio->uio_resid);

#if 0
	/*
	 * If the user buffer is smaller than the size of one dos directory
	 * entry or the file offset is not a multiple of the size of a
	 * directory entry, then we fail the read.
	 */
	off = offset = uio->uio_offset;
	if (uio->uio_resid < sizeof(struct direntry) ||
	    (offset & (sizeof(struct direntry) - 1)))
	  return (EINVAL);
#endif
   
	
	
	_buf = &fent;
	dirent = &dp;
	printf("gfs_readdir(): vp %p, uio %p, cred %p, eofflagp %p ncookies %u\n",
		   ap->a_vp, uio, ap->a_cred, ap->a_eofflag,ap->a_ncookies);

	if ((error = gfs_glock(sdp, dip->i_di.di_glock, dip, GL_SHARED))) {
	  IFDEBUG(DEBUG_FILE1,
			  printk(DEBUG_HOW "Exiting gfs_readdir()\n"););
	  return(error);
	}

	/* 	printf("\tcalling gfs_dir_trans %p\n",dirent);*/
	error = gfs_dir_trans(dip, &f_pos, _buf, &fake_filldir);

	pos = uio->uio_offset;
	/* 	reclen = sizeof(struct dirent); */
   
	bzero(&dp,sizeof(struct dirent));
	dp.d_namlen = 0;
	/*	dp.d_type = DT_REG; */
	/* 	dp.d_fileno = 1; */
	reclen =  GENERIC_DIRSIZ(&dp);
	dp.d_reclen = 0;
	error = uiomove((caddr_t) &dp, reclen, uio);

	pos += reclen;
	uio->uio_offset = pos;

	IFDEBUG(DEBUG_FILE1,
			printk(DEBUG_HOW "Exiting gfs_readdir()\n"););
 out:
	gfs_gunlock(sdp, dip->i_di.di_glock, 0);
	printf("gfs_readdir uio->uio_offset %u uio->uio_resid %u\n",(u_int)uio->uio_offset,uio->uio_resid);
/* 	Debugger("gfs_readdir"); */
	return (error);
}

/*
 * Return target name of a symbolic link
 */
int
gfs_readlink(ap)
	struct vop_readlink_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
	} */ *ap;
{
#if 0
	register struct vnode *vp = ap->a_vp;
	int isize;
#endif
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Entering gfs_readlink()\n"););
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_readlink(OK)\n"););
	return (EINVAL);
}

/*
 * Create a regular file
 */
int
gfs_create(ap)
	struct vop_create_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap;
{


  struct vnode *dvp = ap->a_dvp;
  struct vnode **vpp = ap->a_vpp;
  struct componentname *cnp = ap->a_cnp;
  gfs_inode_t *dip = vtogi(dvp);
  int error = 0;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct qstr str;
  uint64 inode_num;

  int mode =  ap->a_vap->va_mode;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_create()\n"););

  str.name = cnp->cn_nameptr;
  str.len = strlen(cnp->cn_nameptr);
 restart:
	
	error = gfs_createi(dip, &str, 0, mode, &inode_num, 0);
	printf ("did the create got num: %u\n",(u_int)inode_num);
	if (error) {
	  IFDEBUG(DEBUG_INODE1,
			  printk(DEBUG_HOW "Exiting gfs_create(ERROR)\n"););
	  return(error);
	}
	
	
	error = gfs_lookupi(dip, &str, vpp, 0);
	if (error)
	  {
		gfs_gunlock(sdp, gfs_ino2gl(sdp, inode_num), 0);
		gfs_gunlock(sdp, dip->i_di.di_glock, 0);
		
		if (error == -EDEADLK)
		  goto restart;
		
		IFDEBUG(DEBUG_INODE1,
				printk(DEBUG_HOW "Exiting gfs_create(ERROR)\n"););
		return(error);    
	  }
	
	
	/*	gfs_gunlock(sdp, vi2ino(inode)->i_di.di_glock, 0); */
	/*	gfs_gunlock(sdp, vtogi(*vpp)->i_di.di_glock, 0); */
	/* added check with ken to see if this correct RMC */
	gfs_gunlock(sdp, gfs_ino2gl(sdp, inode_num), 0);
	gfs_gunlock(sdp, dip->i_di.di_glock, 0);
	IFDEBUG(DEBUG_SUPER1,
			printk(DEBUG_HOW "Exiting gfs_create(OK)\n"););
	return (0);
}



static int
gfs_lookup(ap)
	 struct vop_lookup_args /* {
							   struct vnode * a_dvp;
							   struct vnode ** a_vpp;
							   struct componentname * a_cnp;
							   } */ *ap;
{
  struct vnode **vpp = ap->a_vpp;
  struct componentname *cnp = ap->a_cnp;
  struct vnode *dvp = ap->a_dvp;

  int flags = cnp->cn_flags;
  int nameiop = cnp->cn_nameiop;

  gfs_inode_t *dip = vtogi(dvp);
#if 0
  gfs_dcached_t *dcd;
#endif
  struct qstr str;
  long int s;

  int error = 0;

  *vpp = NULL;

  IFDEBUG(DEBUG_INODE1,
		  printf("Entering gfs_lookup()\n");
		  printf("  Name: %s\n", cnp->cn_nameptr);
		  printf("gfs_lookupi cnp->cn_nameiop: %lu\n", cnp->cn_nameiop);
		  printf("  Flags: %u\n", (uint)cnp->cn_flags););


  str.name = cnp->cn_nameptr;
  str.len = strlen(cnp->cn_nameptr);

  switch (nameiop){
  case CREATE:
	printf("gfs_lookup nameiop CREATE\n");
	break;
  case RENAME:
	printf("gfs_lookup nameiop RENAME\n");
	break;
  case LOOKUP:
	printf("gfs_lookup nameiop LOOKUP\n");
	break;
  case DELETE:
	printf("gfs_lookup nameiop DELETE\n");
	break;
  default:
	printf("gfs_lookup nameiop %d\n",nameiop);
	break;
  }
  if (flags & ISLASTCN){
	printf("gfs_lookup flags & ISLASTCN\n");
  }
  
  if ((nameiop == CREATE || nameiop == RENAME) &&
	  (flags & ISLASTCN)) {
	/* under ufs there is bunch of slot shit?... probably for cache stuff */
	/* I'm going to ignore that since I don't thing gfs can cache dir stuff anyways */
	IFDEBUG(DEBUG_INODE1,
			printk(DEBUG_HOW "Exiting gfs_lookup(EJUSTRETURN)\n"););
	cnp->cn_flags |= SAVENAME;
	/*	Debugger("gfs_lookup"); */
	return (EJUSTRETURN);
  }
	

 restart:
  error = gfs_lookupi(dip, &str, vpp, 0);

  if(error == -EDEADLK){
	 IFDEBUG(DEBUG_DEADLOCK,
	    printk(DEBUG_HOW "  lookup/clear_inode deadlock, retrying...\n"););
	 /*    current->state = TASK_INTERRUPTIBLE; */
	 /*    schedule_timeout(HZ >> 3); */
	 tsleep(&s, PZERO, "narf", hz >> 3);
	 goto restart;
  } else if (error && error != -ENOENT)  {
    IFDEBUG(DEBUG_INODE2,
			printk(DEBUG_HOW "  error = %d\n", error););
	
    IFDEBUG(DEBUG_INODE1,
			printk(DEBUG_HOW "Exiting gfs_lookup(ERROR)\n"););
    return (-error); /* this sign shit on the errors is going to drive me nuts */
  }
  
  IFDEBUG(DEBUG_INODE1,
		  printk(DEBUG_HOW "Exiting gfs_lookup(OK)\n"););
  return (-error);
}

int
gfs_setattr(ap)
	 struct vop_setattr_args /* {
								struct vnodeop_desc *a_desc;
								struct vnode *a_vp;
								struct vattr *a_vap;
								struct ucred *a_cred;
								struct proc *a_p;
								} */ *ap;
{
  struct vnode *vp = ap->a_vp;
  struct vattr *vap = ap->a_vap;
  
  printf("Entering gfs_setattr\n");

  if ((vap->va_flags != VNOVAL || 
	   vap->va_uid != (uid_t)VNOVAL ||
	   vap->va_gid != (gid_t)VNOVAL ||
	   vap->va_atime.tv_sec != VNOVAL ||
	   vap->va_mtime.tv_sec != VNOVAL ||
	   vap->va_mode != (mode_t)VNOVAL) &&
	  (vp->v_mount->mnt_flag & MNT_RDONLY))
	return (EROFS);
  if (vap->va_size != VNOVAL) {
	switch (vp->v_type) {
	case VDIR:
	  return (EISDIR);
	case VCHR:
	case VBLK:
	case VSOCK:
	case VFIFO:
	  if (vap->va_flags != VNOVAL)
		return (EOPNOTSUPP);
	  return (0);
	case VREG:
	case VLNK:
	default:
	  /*
	   * Disallow write attempts if the filesystem is
	   * mounted read-only.
	   */
	  if (vp->v_mount->mnt_flag & MNT_RDONLY)
		return (EROFS);
	}
  }
  printf("Exiting gfs_setattr\n");
  return (0);
}

/*
 *  We handle getattr only to change the fsid.
 */
static int
gfs_getattr(ap)
	 struct vop_getattr_args /* {
								struct vnode *a_vp;
								struct vattr *a_vap;
								struct ucred *a_cred;
								struct proc *a_p;
								} */ *ap;
{
#if 0
  int error;
#endif
  gfs_inode_t *ip = ap->a_vp->v_data;
  struct vattr *va = ap->a_vap;

  IFDEBUG(DEBUG_INODE1,
		printf(DEBUG_HOW "Entering gfs_getattr()\n"););

  VATTR_NULL(va); /* start fresh */


  switch (ip->i_di.di_type)
	{
	case FILE_REG:
	  va->va_type = VREG;
	  break;
	case FILE_DIR:
	  va->va_type = VDIR;
	  break;
	case FILE_LNK:
	  va->va_type = VLNK;
	  break;
	default:
	  printk(KERN_ERR "GFS:  D'oh!  dinode %u/%u has bad type %d\n",
			 (uint)(ip->i_num >> 32),(uint)ip->i_num, ip->i_di.di_type);
	  break;
	};

  va->va_fsid = 0; /* not sure what to set this to, device number of device I think */
  va->va_fileid = 0;
  va->va_flags = 0;
  
  va->va_nlink = ip->i_di.di_nlink;
  va->va_uid = ip->i_di.di_uid;
  va->va_gid = ip->i_di.di_gid;


#ifdef __alpha__
  va->va_size = ip->i_di.di_bsize;
#else 
  va->va_size = (ip->i_di.di_bsize & 0xFFFFFFFF80000000ll) ? -1 : ip->i_di.di_bsize;
#endif  /*  __alpha__  */
  
  /* do we store nsec values anywhere? is it even part of the dinode */
  va->va_atime.tv_sec = ip->i_di.di_atime;
  va->va_atime.tv_nsec = 0; 
  va->va_mtime.tv_sec = ip->i_di.di_mtime;
  va->va_mtime.tv_nsec = 0;
  va->va_ctime.tv_sec = ip->i_di.di_ctime;
  va->va_ctime.tv_nsec = 0;

  va->va_blocksize = 1 << ip->i_sbd->sd_fsb2bb_shift;
  /*  va->va_blocksize = 4096; */

  va->va_bytes = dbtob((u_quad_t)ip->i_di.di_blocks);

  va->va_mode = ip->i_di.di_mode;  /* not sure this is that correct */
  
  /*  Debugger("gfs_getattr"); */
  IFDEBUG(DEBUG_INODE1,
          printf(DEBUG_HOW "Exiting gfs_getatttr()\n"););
  return (0);
}

static int
gfs_access(ap)
	 struct vop_access_args /* {
							   struct vnode *a_vp;
							   int  a_mode;
							   struct ucred *a_cred;
							   struct proc *a_p;
							   } */ *ap;
{
  struct vnode *vp = ap->a_vp;
  mode_t mode = ap->a_mode;

  printf("Entering gfs_access()\n");
  /*
   * Disallow write attempts on read-only layers;
   * unless the file is a socket, fifo, or a block or
   * character device resident on the file system.
   */
  if (mode & VWRITE) {
	switch (vp->v_type) {
	case VDIR:
	case VLNK:
	case VREG:
	  if (vp->v_mount->mnt_flag & MNT_RDONLY)
		return (EROFS);
	  break;
	default:
	  break;
	}
  }

  printf("Exiting gfs_access()\n");
  return (0);
}

/*
 * We need to process our own vnode lock and then clear the
 * interlock flag as it applies only to our vnode, not the
 * vnodes below us on the stack.
 */
static int
gfs_lock(ap)
	 struct vop_lock_args /* {
							 struct vnode *a_vp;
							 int a_flags;
							 struct proc *a_p;
							 } */ *ap;
{
  int error = 0;
  printf("Entering gfs_lock()\n");
  /*  error = vop_stdlock(ap); */
  printf("Exiting gfs_lock(OK)\n");
  return (error);
}

/*
 * We need to process our own vnode unlock and then clear the
 * interlock flag as it applies only to our vnode, not the
 * vnodes below us on the stack.
 */
static int
gfs_unlock(ap)
	 struct vop_unlock_args /* {
							   struct vnode *a_vp;
							   int a_flags;
							   struct proc *a_p;
							   } */ *ap;
{
  int error = 0;
  printf("Entering gfs_unlock()\n");
  /*  error = vop_stdunlock(ap); */
  printf("Exiting gfs_unlock(OK)\n");
  return (error);
}

static int
gfs_inactive(ap)
	 struct vop_inactive_args /* {
								 struct vnode *a_vp;
								 struct proc *a_p;
								 } */ *ap;
{
  struct vnode *vp = ap->a_vp;
  gfs_inode_t *ip = vp->v_data;
  gfs_sbd_t *sdp = ip->i_sbd;
  struct proc *p = ap->a_p;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_inactive()\n"););
	
  if (prtactive && vp->v_usecount != 0)
	vprint("gfs_inactive: pushing active", vp);
  
  printf("ap: %p v_data %p\n",ap,ip);

  if (!ap || !(ip = (ap->a_vp)->v_data)) {
    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting gfs_inactive(ERROR)\n"););
	Debugger("gfs_inactive");
    return(EINVAL);
  }

  if (!(sdp = ip->i_sbd))  {
    printk(KERN_ERR "GFS:  No superblock when clearing inode %u \n",(uint)ip->i_num);
	/*    goto cleanup;*/
	return(EINVAL);
  } 


  GFS_SEMA_DOWN(&sdp->sd_id_mutex);

#if 0
  /* this is fucked... causes loop in lookupi... why do this anyways */
  ip->i_next = sdp->sd_id_ip;
  sdp->sd_id_ip = ip;
#endif

  sdp->sd_id_count++;

  GFS_SEMA_UP(&sdp->sd_id_mutex);

  GFS_SEMA_UP(&sdp->sd_id_work);

#if 0
 cleanup:
   vi2ino(inode) = NULL;
#endif
   /*    Debugger("gfs_inactive"); */
   VOP_UNLOCK(vp, 0, p);

   if (vp->v_usecount == 0){
#if 0
	 /* this gets done be reclaim... which I think is called by vrecycle */
	 gfree(vp->v_data,0); /* size not used in bsd */
	 vp->v_data = NULL;
#endif
	 vrecycle(vp, (struct simplelock *)0, p);
   }
   
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_inactive(OK)\n"););

  return (0);
}

/*
 * _reclaim is called when getnewvnode()
 * wants to make use of an entry on the vnode
 * free list.  at this time the filesystem needs
 * to free any private data and remove the node
 * from any private lists.
 */

static int
gfs_reclaim(ap)
	 struct vop_reclaim_args /* {
								struct vnode *a_vp;
								struct proc *a_p;
								} */ *ap;
{
  struct vnode *vp = ap->a_vp;
  gfs_inode_t *ip = vp->v_data;


  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_reclaim()\n"););

  gfs_ihashdel(ip);
  
  cache_purge(vp);
  if (ip->i_vino) {
	/*	vrele(ip->i_vino); */ /* already at 0 ... hmm why */
	ip->i_vino = 0;
  }
  gfree(ip,0);
  vp->v_data = NULL;
  
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_reclaim(OK)\n"););
  return (0);
}

static int
gfs_print(ap)
	 struct vop_print_args /* {
							  struct vnode *a_vp;
							  } */ *ap;
{

  register struct vnode *vp = ap->a_vp;
  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_print()\n"););
  printf ("\ttag VT_GFS, vp=%p, lowervp=%p\n", vp, GFSVPTOLOWERVP(vp));

  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_print(OK)\n"););
  return (0);
}

/*
 * XXX - vop_strategy must be hand coded because it has no
 * vnode in its arguments.
 * This goes away with a merged VM/buffer cache.
 */
static int
gfs_strategy(ap)
	 struct vop_strategy_args /* {
								 struct buf *a_bp;
								 } */ *ap;
{
  struct buf *bp = ap->a_bp;
  int error;
  struct vnode *savedvp;

  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Entering gfs_strategy()\n"););

  savedvp = bp->b_vp;
  bp->b_vp = GFSVPTOLOWERVP(bp->b_vp);

  error = VOP_STRATEGY(bp->b_vp, bp);

  bp->b_vp = savedvp;

  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_strategy(OK)\n"););
  return (error);
}

/*
 * XXX - like vop_strategy, vop_bwrite must be hand coded because it has no
 * vnode in its arguments.
 * This goes away with a merged VM/buffer cache.
 */
#if 0
static int
gfs_bwrite(ap)
	 struct vop_bwrite_args /* {
							   struct buf *a_bp;
							   } */ *ap;
{
  struct buf *bp = ap->a_bp;
  int error;
  struct vnode *savedvp;

  IFDEBUG(DEBUG_SUPER1,
		printk(DEBUG_HOW "Entering gfs_bwrite()\n"););

  savedvp = bp->b_vp;
  bp->b_vp = GFSVPTOLOWERVP(bp->b_vp);

  error = VOP_BWRITE(bp);

  bp->b_vp = savedvp;

  IFDEBUG(DEBUG_SUPER1,
		  printk(DEBUG_HOW "Exiting gfs_bwrite(OK)\n"););

  return (error);
}
#endif
/*
 * Global vfs data structures
 */
vop_t **gfs_vnodeop_p;
static struct vnodeopv_entry_desc gfs_vnodeop_entries[] = {
	{ &vop_default_desc,		(vop_t *) vop_defaultop },
	{ &vop_access_desc,			(vop_t *) gfs_access },
	/*	{ &vop_bwrite_desc,			(vop_t *) gfs_bwrite }, */
	{ &vop_getattr_desc,		(vop_t *) gfs_getattr },
	{ &vop_inactive_desc,		(vop_t *) gfs_inactive },
	{ &vop_lock_desc,		    (vop_t *) gfs_lock },
	{ &vop_lookup_desc,			(vop_t *) gfs_lookup },
	{ &vop_print_desc,			(vop_t *) gfs_print },
	{ &vop_reclaim_desc,		(vop_t *) gfs_reclaim },
	{ &vop_setattr_desc,		(vop_t *) gfs_setattr },
	{ &vop_strategy_desc,		(vop_t *) gfs_strategy },
	{ &vop_unlock_desc,			(vop_t *) gfs_unlock },
	{ &vop_create_desc,     	(vop_t *) gfs_create },
	{ &vop_fsync_desc,		    (vop_t *) gfs_missingop },
	{ &vop_read_desc,		    (vop_t *) gfs_missingop },
	{ &vop_reallocblks_desc,	(vop_t *) gfs_missingop },
	{ &vop_write_desc,		    (vop_t *) gfs_missingop },
	{ &vop_abortop_desc,		(vop_t *) gfs_abortop },
	{ &vop_advlock_desc,		(vop_t *) gfs_advlock },
	{ &vop_bmap_desc,		    (vop_t *) gfs_bmap },
	{ &vop_cachedlookup_desc,	(vop_t *) gfs_missingop },
	{ &vop_close_desc,		    (vop_t *) gfs_close },
	{ &vop_islocked_desc,		(vop_t *) vop_stdislocked },
	{ &vop_link_desc,		    (vop_t *) gfs_link },
	{ &vop_mkdir_desc,			(vop_t *) gfs_mkdir },
	{ &vop_mknod_desc,			(vop_t *) gfs_mknod },
	{ &vop_mmap_desc,			(vop_t *) gfs_mmap },
	{ &vop_open_desc,			(vop_t *) gfs_open },
	{ &vop_pathconf_desc,		(vop_t *) gfs_pathconf },
	{ &vop_poll_desc,			(vop_t *) vop_stdpoll },
	{ &vop_readdir_desc,		(vop_t *) gfs_readdir },
	{ &vop_readlink_desc,		(vop_t *) gfs_readlink },
	{ &vop_remove_desc,			(vop_t *) gfs_remove },
	{ &vop_rename_desc,			(vop_t *) gfs_rename },
	{ &vop_rmdir_desc,			(vop_t *) gfs_rmdir },
	{ &vop_symlink_desc,		(vop_t *) gfs_symlink },
	{ &vop_whiteout_desc,		(vop_t *) gfs_whiteout },
	{ NULL, NULL }
};

static struct vnodeopv_desc gfs_vnodeop_opv_desc =
{ &gfs_vnodeop_p, gfs_vnodeop_entries };

VNODEOP_SET(gfs_vnodeop_opv_desc);
