/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: gfs.h
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_DOT_H__
#define __GFS_DOT_H__



/*  OS specific files  */

#if defined(__linux__)
#include "gfs_linux.h"
#elif defined(__FreeBSD__)
#include "gfs_fbsd.h"
#endif



/*  Extern Macro  */

#ifndef EXTERN
#define EXTERN extern
#define INIT(X)
#else
#undef EXTERN
#define EXTERN
#define INIT(X) =X 
#endif


/*  Includes  */

#include "GFS.cf"
#include "global.h"
#include "gfs_ondisk.h"
#include "incore.h"
#include "gfs_debug.h"
#include "gfs_profile.h"


/*  Macros  */

#ifdef ASSERT
#undef ASSERT
#endif  /*  ASSERT  */

#ifdef DEBUG
#define ASSERT_DEBUG            \
                         int y; \
                                \
                         for (y=0; y < stack_ct; y++)  \
                            printk("stack %d: %s\n", y, fn_names[stack[y]]);
#else
#define ASSERT_DEBUG
#endif

#define ASSERT(x) {if (!(x)) { \
                         ASSERT_DEBUG \
			 printk(KERN_EMERG "\n\nYou've found a bug in GFS!\n\n"); \
			 printk(KERN_EMERG "Please copy down the following line and send it to gfs-devel@lcse.umn.edu\n\n"); \
			 panic("GFS:  Assertion \"%s\" failed on line %d of file %s\n", #x, \
			       __LINE__, __FILE__); } }

#define GARBAGE_BLOCKS() {printk(KERN_ERR "GFS:  garbage blocks generated by %s on line %d\n", \
				 __FILE__, __LINE__);};



#endif  /*  __GFS_DOT_H__  */
