/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: glock.h
**
**    Description:
**	This module implements the GFS interface to the global locks.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_GLOCK_DOT_H__
#define __GFS_GLOCK_DOT_H__

#include "gfs.h"


/*  Flags passed in to gfs_glock()  */

#define GL_SHARED     (0x00010000)

/*  Flags passed in to gfs_gunlock()  */

#define GL_INCR       (0x00000001)
#define GL_ZAP        (0x00000002)
#define GL_SYNC       (0x00000004)


typedef struct gfs_lockop_s
{
  uint64 lock;
  gfs_inode_t *ip;
  int flags;
} gfs_lockop_t;


/* Functions used by the whole filesystem */

int          gfs_glock(gfs_sbd_t *sdp, uint64 lock, gfs_inode_t *ip, int flags);
int          gfs_gunlock(gfs_sbd_t *sdp, uint64 lock, int flags);
int          gfs_glock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops);
int          gfs_gunlock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops);

/*  Functions used by the dio routines  */

gfs_glock_t *gfs_get_glstruct(gfs_sbd_t *sdp, uint64 lock, int create);
void         gfs_put_glstruct(gfs_sbd_t *sdp, gfs_glock_t *gl);

/*  Only used in gfs_read_super()  */

void         gfs_glock_cb(void *fsdata, uint64 lock, int flags);

/*  Only used during gfs_put_super()  */

int          gfs_demote_glock(gfs_sbd_t *sdp, gfs_glock_t *gl);


#endif /*  __GFS_GLOCK_DOT_H__  */
