/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: lockops.c
**
*******************************************************************************
******************************************************************************/

#define __NO_VERSION__ 
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "gfs_locking.h"
#define DLOCK_VERSION 3
#include "pool.h"

#include "dlock094.h"
#include "udp.h"



#define REPLY_LEN (256)

typedef struct lock_s
{
  uint32 lock;
  unsigned char buf[sizeof(pool_dlock_t) + REPLY_LEN];
  uint32 vn;
  int time;
} lock_t;



dlock094_t *list = NULL;





/******************************************************************************
*******************************************************************************
**
** atoi()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int atoi(char *c)
{
  int x = 0;

  while ('0' <= *c && *c <= '9')
  {
    x = x * 10 + (*c - '0');
    c++;
  }

  return(x);
}


/******************************************************************************
*******************************************************************************
**
** inet_aton()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int inet_aton(char *ascii, uint32 *number)
{
  uint32 ip = 0, value;
  int x;


  for (x = 0; x < 4; x++)
  {
    value = atoi(ascii);
    if (value > 255)
      return(-1);

    ip = (ip << 8) | value;

    if (x != 3)
    {
      ascii = strstr(ascii, ".");
      if (!ascii)
	return(-1);
      ascii++;
    }
  }
	

  *number = cpu_to_be32p(&ip);

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** fill_cdb()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void fill_cdb(unsigned char *buf, int action, uint32 ln, uint32 clientid, int alloc_len)
{
  buf[0] =  0x83;

  buf[1] =  action & 0x0F;

  buf[2] =  (ln >> 24) & 0xFF;
  buf[3] =  (ln >> 16) & 0xFF;
  buf[4] =  (ln >> 8) & 0xFF;
  buf[5] =  ln & 0xFF;

  buf[6] =  (clientid >> 24) & 0xFF;
  buf[7] =  (clientid >> 16) & 0xFF;
  buf[8] =  (clientid >> 8) & 0xFF;
  buf[9] =  clientid & 0xFF;

  buf[10] = (alloc_len >> 24) & 0xFF;
  buf[11] = (alloc_len >> 16) & 0xFF;
  buf[12] = (alloc_len >> 8) & 0xFF;
  buf[13] = alloc_len & 0xFF;

  buf[14] = 0x00;

  buf[15] = 0x00;
}


/******************************************************************************
*******************************************************************************
**
** unlock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int unlock(dlock094_t *md, uint64 lock, lock_t *data, int incr)
{
  pool_dlock_t *dlock = (pool_dlock_t *)data->buf;
  unsigned char *ret_data = &data->buf[sizeof(pool_dlock_t)];

  int x, holders;
  uint32 clientid;
  struct timeval tv;
  uint32 vn;
  int error;


  memset(ret_data, 0, REPLY_LEN);
  fill_cdb(dlock->dl_cdb, (incr) ? 6 : 5, lock, md->clientid, REPLY_LEN);

  dlock->dl_ln_not_device = TRUE;
  dlock->dl_ln = lock;
  dlock->dl_lnoffset = 2;
  dlock->dl_lnsize = 4;
  dlock->dl_alloc_len = REPLY_LEN;

  error = pool_ioctl(md->inode, NULL, POOL_DLOCKOP, (unsigned long)dlock);
  if (error)
    return(error);

  if (!(ret_data[4] & 0x80))
    return(-EIO);

  vn = ((ret_data[0] << 24) |
	(ret_data[1] << 16) |
	(ret_data[2] << 8)  |
	(ret_data[3]));  
  
  data->vn = vn;

  
  holders = ((ret_data[6] << 8) | ret_data[7]) / 4;  

  for (x = 0; x < holders; x++)
  {
    clientid = ((ret_data[8  + x * 4] << 24) |
		(ret_data[9  + x * 4] << 16) |
		(ret_data[10 + x * 4] << 8) |
		(ret_data[11 + x * 4]));
    
    ASSERT(clientid != md->clientid);
  }


  do_gettimeofday(&tv);
  data->time = tv.tv_sec;


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** kick_holders()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void kick_holders(dlock094_t *md, lock_t *data, uint32 lock, uint32 flags)
{
  unsigned char *ret_data = &data->buf[sizeof(pool_dlock_t)];
  int x, holders;
  uint32 clientid;


  holders = ((ret_data[6] << 8) | ret_data[7]) / 4;
  

  for (x = 0; x < holders; x++)
  {
    clientid = ((ret_data[8  + x * 4] << 24) |
		(ret_data[9  + x * 4] << 16) |
		(ret_data[10 + x * 4] << 8) |
		(ret_data[11 + x * 4]));

    if (clientid != md->clientid)
      send_mess(md, clientid, lock, flags);
    else
    {
      printk(KERN_ERR "dlock094:  found lock %u\n", lock);

      unlock(md, lock, data, TRUE);
      data->vn--;  /*  Make sure no cached data is used  */

      break;
    }
  }
}


/******************************************************************************
*******************************************************************************
**
** dlock094_mount()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void *dlock094_mount(char *table_name, char *host_data, gfs_callback_t *cb, void *fsdata)
{
  dlock094_t *moddata;
  char buf[256], name[GFS_LOCKNAME_LEN + 1];
  char *c;


  MOD_INC_USE_COUNT;


  strcpy(name, table_name);

  c = strstr(name, ":");
  if (!c)
  {
    printk(KERN_ERR "dlock094:  tablename should be \"poolname:portnumber\"\n");
    goto fail;
  }
  *c++ = 0;


  /*  Check to make sure this pool isn't already being used  */

  for (moddata = list; moddata; moddata = moddata->next)
    if (!strcmp(moddata->table_name, name))
    {
      printk(KERN_ERR "dlock094:  pool %s is already in use\n", name);
      goto fail;
    }


  /*  Set up our main structure  */

  moddata = (dlock094_t *)kmalloc(sizeof(dlock094_t), GFP_KERNEL);
  if (!moddata)
  {
    printk(KERN_ERR "dlock094:  out of memory\n");
    goto fail;
  }
  memset(moddata, 0, sizeof(dlock094_t));

  strcpy(moddata->table_name, name);

  moddata->cb = cb;
  moddata->fsdata = fsdata;


  /*  Get a pointer to the pool  */

  strcpy(buf, "/dev/pool/");
  strcat(buf, name);

  moddata->dcache = lookup_dentry(buf, NULL, 0);

  if (!moddata->dcache || !moddata->dcache->d_inode)
  {
    if (moddata->dcache)
      dput(moddata->dcache);
    printk(KERN_ERR "dlock094:  can't open pool %s\n", buf);
    goto fail_kfree;
  }

  moddata->inode = moddata->dcache->d_inode;


  /*  Figure out port number  */

  moddata->port = atoi(c);
  if (!moddata->port)
  {
    printk(KERN_ERR "dlock094:  bad port number\n");
    goto fail_dput;
  }
  cpu_to_be16s(&moddata->port);


  /*  Figure out the client id  */

  if (inet_aton(host_data, &moddata->clientid) < 0)
  {
    printk(KERN_ERR "dlock094:  bad IP address %s\n", host_data);
    goto fail_dput;
  }


  /*  Open socket  */

  if (open_socket(moddata) < 0)
    goto fail_dput;


  /*  Start up recvd  */

  moddata->recvd_lock = MUTEX_LOCKED;

  if (kernel_thread(recvd, moddata, 0) < 0)
    goto fail_closesock;

  down(&moddata->recvd_lock);


  /*  Success  */

  moddata->next = list;
  list = moddata;

  return(moddata);



 fail_closesock:
  close_socket(moddata);

 fail_dput:
  dput(moddata->dcache);

 fail_kfree:
  kfree(moddata);

 fail:
  MOD_DEC_USE_COUNT;
  return(NULL);
}


/******************************************************************************
*******************************************************************************
**
** dlock094_unmount()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dlock094_unmount(void *moddata)
{
  dlock094_t *tmp, *last = NULL, *md = (dlock094_t *)moddata;
  int done = FALSE;


  if (list == md)
  {
    list = list->next;
    done = TRUE;
  }
  else
    for (tmp = list; tmp; last = tmp, tmp = tmp->next)
      if (tmp == md)
      {
	last->next = tmp->next;
	done = TRUE;
      }


  if (!done)
    printk(KERN_ERR "dlock094:  given a weird moddata pointer\n");
  else
  {
    md->recvd_stop = TRUE;
    send_mess(md, md->clientid, 0, 0);
    down(&md->recvd_lock);

    close_socket(md);
    dput(md->dcache);
    kfree(md);
  }

  MOD_DEC_USE_COUNT;
}


/******************************************************************************
*******************************************************************************
**
** dlock094_lock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int dlock094_lock(void *moddata, uint64 lock, int op, int flags, void **lockdata)
{
  dlock094_t *md = (dlock094_t *)moddata;

  lock_t *data;
  pool_dlock_t *dlock;
  unsigned char *ret_data;

  int cont = TRUE;
  int error, ret = -1;
  uint32 vn;
  int new;
  int tries = 0;


  if (*lockdata)
  {
    data = (lock_t *)*lockdata;
    new = FALSE;
  }
  else
  {
    data = (lock_t *)kmalloc(sizeof(lock_t), GFP_KERNEL);
    if (!data)
      return(-ENOMEM);
    memset(data, 0, sizeof(lock_t));
    data->lock = lock;
    *lockdata = data;
    new = TRUE;
  }

  ASSERT(data->lock == lock);

  dlock = (pool_dlock_t *)data->buf;
  ret_data = &data->buf[sizeof(pool_dlock_t)];


  ASSERT(op != LCK_OP_TEST);
  

  if ((flags & LCK_COMMUTE_MOD) || ((flags & LCK_COMMUTE) && !(flags & LCK_SHARED)))
  {
    ASSERT(!new);

    error = unlock(md, lock, data, (flags & LCK_COMMUTE_MOD) ? TRUE : FALSE);
    if (error < 0)
      return(error);
  }


  do
  {
    memset(ret_data, 0, REPLY_LEN);
    fill_cdb(dlock->dl_cdb, (flags & LCK_SHARED) ? 1 : 2, lock, md->clientid, REPLY_LEN);

    dlock->dl_ln_not_device = TRUE;
    dlock->dl_ln = lock;
    dlock->dl_lnoffset = 2;
    dlock->dl_lnsize = 4;
    dlock->dl_alloc_len = REPLY_LEN;

    error = pool_ioctl(md->inode, NULL, POOL_DLOCKOP, (unsigned long)dlock);
    if (error)
      return(error);

    if (ret_data[4] & 0x80)
    {
      ret = LCKOUT_HELD;

      vn = ((ret_data[0] << 24) |
	    (ret_data[1] << 16) |
	    (ret_data[2] << 8)  |
	    (ret_data[3]));  

      if (!new && vn == data->vn)
      {
	struct timeval tv;
	do_gettimeofday(&tv);
	if (tv.tv_sec - data->time < 3600)
	  ret |= LCKOUT_CACHEABLE;
      }

      data->vn = vn;

      switch (ret_data[4] & 0x03)
      {
      case 0:
	printk(KERN_ERR "dlock094:  Whoah!  Unlocked state after acquire\n");
	ASSERT(0);
	break;
      case 1:
	ASSERT(flags & LCK_SHARED);
	ret |= LCKOUT_SHARED;
	break;
      case 2:
	ASSERT(!(flags & LCK_SHARED));
	break;
      default:
	printk(KERN_ERR "dlock094:  weird state value\n");
	break;
      }

      switch ((ret_data[4] >> 2) & 0x03)
      {
      case 0:
	break;
      case 1:
	ret |= LCKOUT_EXPIRED_S;
	break;
      case 2:
	ret |= LCKOUT_EXPIRED_E;
	break;
      default:
	printk(KERN_ERR "dlock094:  weird expired value\n");
	break;
      }

      cont = FALSE;
    }
    else
    {
      if (op == LCK_OP_ACQUIRE)
      {
	if (tries % 100 == 0)
	  kick_holders(md, data, lock, (flags & LCK_SHARED) ? CB_NEED_S : CB_NEED_E);
	if (tries % 1000 == 999)
	  printk(KERN_ERR "dlock094:  asking for lock %"FMT_64"u %s\n",
		 lock, ((flags & LCK_SHARED) ? "shared" : "exclusive"));
	tries++;
      }
      else if (op == LCK_OP_TRY)
      {
	ret = LCKOUT_CACHEABLE;
	cont = FALSE;
      }
    }
  } while (cont);


  return(ret);
}


/******************************************************************************
*******************************************************************************
**
** dlock094_unlock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int dlock094_unlock(void *moddata, uint64 lock, int flags, void **lockdata)
{
  dlock094_t *md = (dlock094_t *)moddata;
  lock_t *data = (lock_t *)*lockdata;

  int error;


  ASSERT(data);
  ASSERT(data->lock == lock);


  error = unlock(md, lock, data, (flags & ULCK_MODIFIED) ? TRUE : FALSE);


  return(error);
}


/******************************************************************************
*******************************************************************************
**
** dlock094_release()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dlock094_release(void *moddata, void **lockdata)
{
  kfree(*lockdata);
  *lockdata = NULL;
}


