/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_ip.o
*
*/
#ifndef __dlock_ipd_h__
#define __dlock_ipd_h__

#include "gfs_locking.h"

/* debug macros */
#undef printd             /* undef it, just in case */
#ifdef DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define printd(fmt, args...) printk( KERN_DEBUG "dlip: " fmt, ## args)
#  else
     /* This one for user space */
#    define printd(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define printd(fmt, args...) /* not debugging: nothing */
#endif

#undef printdd
#define printdd(fmt, args...) /* nothing: it's a placeholder */

#ifdef ASSERT
#undef ASSERT
#endif  /*  ASSERT  */
#define ASSERT(x) {if (!(x)) { \
	printk(KERN_EMERG "\n\nYou've found a bug in dlock_ip!\n\n"); \
	printk(KERN_EMERG "Please copy down the following line and send it to gfs-devel@lcse.umn.edu\n\n"); \
	panic("dlock_ip:  Assertion \"%s\" failed on line %d of file %s\n", #x, \
			 __LINE__, __FILE__); } }

/******************************************************************************
 * This structure builds a list of the mounted lock tables.  Gfs grabs one
 * table per filesystem mounted.
 *****************************************************************************/
typedef struct dlock_ip_s
{
  char table_name[GFS_LOCKNAME_LEN];

  gfs_callback_t *cb;
  void *fsdata;

  u32 mycid;
  u32 serv_ip;
  u8  maxclnts;
  struct rpc_clnt *clnt;

  struct dlock_ip_s *next;
} dlock_ip_t;
/* May have to move the rpc clnt into the lock data in order to allow
 * multiple connections with multiple locks.  Should not be that difficult
 * to implement.  So I made it a compile time option.
 */
typedef struct lock_s {
#ifdef RPC_PER_LOCK
	struct rpc_clnt *clnt;
#endif
	u32 lock;
	u32 vn;
	int time;
	u8  cidcnt;
	u32 *cids;
} lock_t;

/* Since I was ending up with a mess of parameters I `cleaned' things up
 * with another structure.
 */
typedef struct {
	u8   res;
	u32  vn;
	u8   exp;
	u8   state;
} dlock_return_t;
#endif

