/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_ip.o
*
* This is an attempt to actually use dlock 9.4
* And so far it looks like it works.
*
* *sigh*, this is not looking like an easy addition.  So I am really
* thinking of taking the clienttalk stuff from dlock-0.9.4 and mostly
* pasting it into here.  Which will be a rather crude combination, but it
* should work.  This is yet another reason to rewrite the commlayer.  Which
* is what I'll have to do inorder to get any nice improvements.
*
******************************************************************************/

#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/smp_lock.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "gfs_locking.h"
#include "dlip.h"
#include "dlrpc.h"
#include "dlip_misc.h"
#include "dlip_cmd_wrapper.h"
#include "dlip_askdrop.h"

MODULE_AUTHOR("Mike Tilstra <tilstra@lcse.umn.edu>");
MODULE_DESCRIPTION("This is a gfs locking module that uses a server over "
"rcp instead of talking with a dlock enabled drive. "
"More info at <http://gfs.lcse.umn.edu>");


/******************************************************************************
 * Load time parameters
 *****************************************************************************/
static int refresh_time = 10;
MODULE_PARM(refresh_time,"i");
#ifdef USE_CALLBACKS
static int ear_port = 14242;
MODULE_PARM(ear_port, "i");
#endif


static char *my_ip = NULL;
static u32 my_cid = 0; /* This is derived from my_ip at modinit */
/* Ok, this is a bit strange, but if you try to pass in just the ip, it
 * will fail because insmod is a dolt and trys to read the ip as a number
 * not a string.  But if you stick single char onto the front of the ip it
 * works.  (ala insmod dlock_ip my_ip=p128.101.182.149 )
 * Though I think I will need to strip that char off before passing it to
 * the aton function.
 */
MODULE_PARM(my_ip, "s");


/******************************************************************************
 * Various globals that this module works on.
 *****************************************************************************/
dlock_ip_t *dlock_ip_list = NULL;
static int materia_intact = FALSE;

/******************************************************************************
 * dlip_materia()
 * Sorta a life thread of this module.
 * Right now it just does a refresh for each of the existing tables.  Its
 * going to get more in the future.
 *****************************************************************************/
int dlip_materia(void *data)
{
	struct semaphore *mux = (struct semaphore *)data;
	unsigned long flags;
	dlock_ip_t *tmp;

	MOD_INC_USE_COUNT;

	lock_kernel();

	exit_mm(current);

	current->session = 1;
	current->pgrp = 1;

	spin_lock_irqsave(&current->sigmask_lock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);

	current->fs->umask = 0;

	sprintf(current->comm, "dlip_materia");

	unlock_kernel();
	/*  Wake up the process who started us  */
	up(mux);

	while (1)
	{
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(refresh_time * HZ);

		if( dlock_ip_list == NULL )
			break;

		for(tmp = dlock_ip_list; tmp; tmp = tmp->next)
			do_refreshAll(tmp->clnt, tmp->mycid);
	}

#ifdef DEBUG
	printk("dlock_ip: dlip_materia ceasing to exist.\n");
#endif
	materia_intact = FALSE;
	MOD_DEC_USE_COUNT;
	return 0;
}


/******************************************************************************
 * dlock_ip_mount()
 *  table_name should be the ip of the dlock server.
 *****************************************************************************/

void *dlock_ip_mount(char *table_name, char *host_data, gfs_callback_t *cb, void *fsdata)
{
	dlock_ip_t *moddata;
	struct rpc_clnt *clnt;
	int err;

	MOD_INC_USE_COUNT;

	for (moddata = dlock_ip_list; moddata; moddata = moddata->next)
		if (!strcmp(moddata->table_name, table_name)) {
			printk(KERN_ERR "dlock_ip: table_exists\n");
			goto fail;
		}

	moddata = (dlock_ip_t *)kmalloc(sizeof(dlock_ip_t), GFP_KERNEL);
	if (!moddata) {
		printk(KERN_ERR "dlock_ip: out of memory\n");
		goto fail;
	}
	memset(moddata, 0, sizeof(dlock_ip_t));

	strcpy(moddata->table_name, table_name);

	moddata->cb = cb;
	moddata->fsdata = fsdata;

	if( inet_aton( table_name, &moddata->serv_ip ) <0) {
		printk(KERN_ERR "dlock_ip: Bad IP number for dlock_ipd server.\n");
		goto memfail;
	}

	moddata->mycid = my_cid;

	if( dlip_create_client( moddata->serv_ip, &clnt ) < 0 ) {
		printk(KERN_ERR "dlock_ip: couldn't open rpc client\n");
		goto memfail;
	}
  
	if(rpciod_up() !=0 ) {
		printk(KERN_WARNING "dlock_ip: couldn't start rpciod!\n");
		goto rpcfail;
	}

	{
		u32 numlock, lock_timeout; /* ignored for now */
		if( (err = do_modepage(clnt, &moddata->maxclnts,
					&numlock, &lock_timeout)) < 0 ) {
			printk(KERN_WARNING "dlock_ip: mode page failed! (%d)\n", err);
			goto rpciodfail;
		}
		printk("dlock_ip: using dlock_ipd server %s\n"
				 "          with %d locks\n"
				 "          %d cids in each list\n"
				 "          and %d ms timeouts\n",
				 moddata->table_name,
				 numlock,
				 moddata->maxclnts,
				 lock_timeout);
	}

	moddata->clnt = clnt;

	{
		struct semaphore mux = MUTEX_LOCKED;

		if(materia_intact == FALSE) {
			err = kernel_thread(dlip_materia, &mux, 0);
			if(err < 0) {
				printk(KERN_WARNING
						"dlock_ip: unable to start materia thread (%d)\n", err);
				goto rpciodfail;
			}
			materia_intact = TRUE;
			down(&mux);
		}
	}

	moddata->next = dlock_ip_list;
	dlock_ip_list = moddata;

	return(moddata);

rpciodfail:
	rpciod_down();
rpcfail:
	rpc_shutdown_client(clnt);
memfail:
	kfree( moddata );
fail:
	MOD_DEC_USE_COUNT;
	return NULL;
}


/******************************************************************************
 * dlock_ip_unmount()
 *****************************************************************************/

void dlock_ip_unmount(void *moddata)
{
	dlock_ip_t *tmp, *last = NULL;
	dlock_ip_t *md = (dlock_ip_t *)moddata;
	int done = FALSE;

	if (dlock_ip_list == moddata) {
		dlock_ip_list = dlock_ip_list->next;
		done = TRUE;
	}
	else
		for (tmp = dlock_ip_list; tmp; last = tmp, tmp = tmp->next)
			if (tmp == moddata) {
				last->next = tmp->next;
				done = TRUE;
			}

	if (!done)
		printk(KERN_ERR "dlock_ip:  given a weird moddata pointer\n");
	else {
		if( md->clnt != NULL )
			rpc_shutdown_client( md->clnt );
		rpciod_down();
		kfree(moddata);
		MOD_DEC_USE_COUNT;
	}


}

/******************************************************************************
 * unlock()
 *****************************************************************************/

static int unlock(dlock_ip_t *md, uint64 lock, lock_t *data, int incr)
{
	struct timeval tv;
	int error,i;
	dlock_return_t rtn;

	printd(" unlocking %d, by %d\n", lock, current->pid);

	error = do_dlock(md, data, ((incr)?6:5), lock, &rtn);
	if (error)
		return(error);
	if(! rtn.res )
		return(-EIO);

	data->vn = rtn.vn;

	/* Walk through cids list and make sure we're not there*/
	for(i=0; i< data->cidcnt; i++) {
		ASSERT( data->cids[i] != md->mycid );
	}

	do_gettimeofday(&tv);
	data->time = tv.tv_sec;

	return(0);
}

/******************************************************************************
 * kick_holders()
 * send a messages to all of the other clients that we want this lock.
 * Unless callbacks are disabled, then it just makes sure we don't have it.
 * (weird case, but it does happen)
 *****************************************************************************/
void kick_holders(dlock_ip_t *md, lock_t *data, u32 lock, u32 flags)
{
	int i;

	for( i = 0; i < data->cidcnt; i++) {
		if( data->cids[i] != md->mycid ) {
#ifdef USE_CALLBACKS
			send_askdrop( data->cids[i], ear_port, md->serv_ip, lock, flags);
#endif
		} else {
			printk(KERN_ERR "dlock_ip: found lock %u\n", lock);
			unlock(md, lock, data, TRUE);
			data->vn--;
			break;
		}
	}
}

/******************************************************************************
 * dlock_ip_lock()
 * Translates gfs's glocks into ipdlocks.
 *****************************************************************************/

int dlock_ip_lock(void *moddata, uint64 lock, int op, int flags,
		void **lockdata)
{
	dlock_ip_t *md = (dlock_ip_t *)moddata;
	lock_t *data;
	dlock_return_t rtn;
	int cont = TRUE;
	int error, ret = -1;
	int new;
	int tries = 0;

	printd(" locking %d, by %d\n", lock, current->pid);

	if (*lockdata) {
		data = (lock_t *)*lockdata;
		new = FALSE;
	} else {
		data = (lock_t *)kmalloc(sizeof(lock_t), GFP_KERNEL);
		if (!data)
			return(-ENOMEM);
		memset(data, 0, sizeof(lock_t));
		if(md->maxclnts != 0) {
			data->cids = (u32*)kmalloc( md->maxclnts * sizeof(u32), GFP_KERNEL);
			if(!data->cids) {
				kfree(data);
				return -EIO;
			}
			memset(data->cids, 0, md->maxclnts * sizeof(u32) );
		}
#ifdef RPC_PER_LOCK
		error = dlip_create_client( md->serv_ip, &data->clnt );
		if(error) {
			kfree(data->cids);
			kfree(data);
			return error;
		}
		if(rpciod_up() != 0) {
			rpc_shutdown_client(data->clnt);
			kfree(data->cids);
			kfree(data);
			return -EIO;
		}
#endif
		data->lock = lock;
		*lockdata = data;
		new = TRUE;
	}

	ASSERT(data->lock == lock );

	ASSERT(op != LCK_OP_TEST); /* Why's this here? */

	if (flags & (LCK_COMMUTE | LCK_COMMUTE_MOD)) {
		ASSERT(!new);

		error = unlock(md, lock, data, (flags & LCK_COMMUTE_MOD) ? TRUE : FALSE);
		if (error < 0)
			return(error);
	}

	do {
		error = do_dlock(md, data, (flags & LCK_SHARED)?1:2, lock, &rtn);
		if (error)
			return(error);

		if ( rtn.res ) {
			ret = LCKOUT_HELD;
#ifndef USE_CALLBACKS
			ret |= LCKOUT_NEED_E;
#endif

			if (!new && rtn.vn == data->vn) {
				struct timeval tv;
				do_gettimeofday(&tv);
				if (tv.tv_sec - data->time < 3600)
					ret |= LCKOUT_CACHEABLE;
			}

			data->vn = rtn.vn;

			switch(rtn.state) {
				case 0:
					printk(KERN_ERR "dlock_ip: Ack! Unlocked state after acquire\n");
					ASSERT(0);
					break;
				case 1:
					ASSERT(flags & LCK_SHARED);
					ret |= LCKOUT_SHARED;
					break;
				case 2:
					ASSERT(!(flags & LCK_SHARED));
					break;
				default:
					printk(KERN_ERR "dlock_ip: weird state value\n");
					break;
			}

			switch(rtn.exp) {
				case 0:
					break;
				case 1:
					ret |= LCKOUT_EXPIRED_S;
					break;
				case 2:
					ret |= LCKOUT_EXPIRED_E;
					break;
				default:
					printk(KERN_ERR "dlock_ip: weird expired value\n");
					break;
			}

			cont = FALSE;
		} else {
			if (op == LCK_OP_ACQUIRE) {
				if (tries % 250 == 0)
					kick_holders(md, data, lock,
							(flags&LCK_SHARED)?CB_NEED_S:CB_NEED_E);
				if (tries % 1000 == 0)
				  printk(KERN_ERR "dlock_ip:  asking for lock %"FMT_64"u\n", lock);

				tries++;
				schedule_timeout(1); /* hopefully give time to other threads*/
			}
			else if (op == LCK_OP_TRY) {
				ret = LCKOUT_CACHEABLE;
				cont = FALSE;
			}
		}
	} while (cont);

	return(ret);
}


/******************************************************************************
 * dlock_ip_unlock()
 * Unlocks a lock.
 *****************************************************************************/

int dlock_ip_unlock(void *moddata, uint64 lock, int flags, void **lockdata)
{
	dlock_ip_t *md = (dlock_ip_t *)moddata;
	lock_t *data = (lock_t *)*lockdata;
	int error;

	ASSERT(data);
	ASSERT(data->lock == lock);

	error = unlock(md, lock, data, (flags & ULCK_MODIFIED) ? TRUE : FALSE);

	return(error);
}


/******************************************************************************
 * dlock_ip_release()
 * cleans up the lockdata when called.
 *****************************************************************************/
void dlock_ip_release(void *moddata, void **lockdata)
{
	lock_t *data = (lock_t*)*lockdata;
#ifdef RPC_PER_LOCK
	rpc_shutdown_client( data->clnt );
	rpciod_down();
#endif
	kfree( data->cids );
	kfree(*lockdata);
	*lockdata = NULL;
}

/******************************************************************************
 *****************************************************************************/
gfs_lock_proto_t dlock_ip_lock_proto = {
	"dlock_ip",
	dlock_ip_mount,
	dlock_ip_unmount,
	dlock_ip_lock,
	dlock_ip_unlock,
	dlock_ip_release,
	NULL
};

/******************************************************************************
 * The beginning and the end of this module.
 *****************************************************************************/
int init_module(void)
{
	int error;

	/* setup cid */
	if( my_ip == NULL ) {
		printk(KERN_ERR "dlock_ip: Must give an IP for my_ip\n"
		                "    You forgot to add my_ip=p#.#.#.#\n");
		return -1;
	}
	printd("ip string = `%s'\n", &my_ip[1]);
	if( (error = inet_aton( &my_ip[1], &my_cid )) < 0) {
		printk(KERN_ERR "dlock_ip: Bad IP number for my_ip\n");
		return error;
	}
	printd("ip num = %x\n",my_cid);
	
	/* fire up the socket stuff */
#ifdef USE_CALLBACKS
	error = init_mastersock(my_cid, ear_port);
	if(error <0) return error;
#endif

	/* load the gfs lock module */
	if ((error = register_lock_proto(&dlock_ip_lock_proto)) < 0)
		return(error);

	setup_rpc_stats();
	printk("dlock_ip: module loaded\n");
	return(0);
}

void cleanup_module(void)
{
#ifdef USE_CALLBACKS
			/* This send is hackery, find a way to fix. */
			send_askdrop(my_cid, ear_port, my_cid, 0, 0);
			close_mastersock();
#endif

	unregister_lock_proto(&dlock_ip_lock_proto);
	shutdown_rpc_stats();
	printk("dlock_ip: module unloaded\n");
}


/*  Symbols  */
#ifndef DEBUG
EXPORT_NO_SYMBOLS;
#endif

