/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_tcp.o
*
*/
#ifndef __dlock_tcp_h__
#define __dlock_tcp_h__

#include "gfs_locking.h"

/* debug macros */
#undef printd             /* undef it, just in case */
#ifdef DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define printd(fmt, args...) printk( KERN_DEBUG "dlock_tcp: " fmt, ## args)
#  else
     /* This one for user space */
#    define printd(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define printd(fmt, args...) /* not debugging: nothing */
#endif

#undef printdd
#define printdd(fmt, args...) /* nothing: it's a placeholder */

/* Easy way to print out errs.*/
#undef printe
#define printe(fmt, args...) printk(KERN_ERR "dlock_tcp: " fmt, ## args)

#ifdef ASSERT
#undef ASSERT
#endif  /*  ASSERT  */
#define ASSERT(x) {if (!(x)) { \
	printk(KERN_EMERG "\n\nYou've found a bug in dlock_tcp!\n\n"); \
	printk(KERN_EMERG "Please copy down the following line and send it to gfs-devel@lcse.umn.edu\n\n"); \
	panic("dlip_tcp:  Assertion \"%s\" failed on line %d of file %s\n", #x, \
			 __LINE__, __FILE__); } }

#define MIN(a,b) ((a<b)?a:b)

/******************************************************************************
 * This structure builds a list of the mounted lock tables.  Gfs grabs one
 * table per filesystem mounted.
 *****************************************************************************/
typedef struct dlock_tcp_s
{
  char table_name[GFS_LOCKNAME_LEN];

  gfs_callback_t *cb;
  void *fsdata;
  
  /* These three are in big endian */
  u32 mycid;
  u32 servip;
  u16 servport;

  /* bunch of stuff that the communication layer needs.
	* Should probably be blocked off into its own struct, but why bother.
	*/
  struct socket *sock;
  struct semaphore sender;
  struct semaphore recver;
  struct semaphore tiodmux;
  int recvd_run;
  char *recv_buf;
  int recv_len;
  int recv_err;

  struct dlock_tcp_s *next;
} dlock_tcp_t;

typedef struct lock_s {
	u32 lock;
	u32 vn;
	int time;
	u8  cidcnt;
	u32 *cids;
} lock_t;

/* Since I was ending up with a mess of parameters I `cleaned' things up
 * with another structure.
 */
typedef struct {
	u8   res;
	u32  vn;
	u8   exp;
	u8   state;
} dlock_return_t;
#endif

