/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_tcp.o
*
* These are just some wrapped up functions for working the comm layer.
*/

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/string.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "dlip.h"
#include "dlip_trans.h"
#include "dlip_data_structures.h"
#include "dlip_cmd_wrapper.h"


/******************************************************************************
 * do_dlock()
 * Builds up the transport structures, and does the rpc_call to the server.
******************************************************************************/
int do_dlock(dlock_tcp_t *md, lock_t *data,
		u8 action, uint64 lock, 
		dlock_return_t *rtn)
{
	dlock_cdb_t in;
	dlock_rdf_t out;
	int err;

	memset(&in, 0, sizeof(dlock_cdb_t));
	memset(&out, 0, sizeof(dlock_rdf_t));

	in.code    = CDBdlock;
	in.action  = action;
	in.ivn     = data->vn;
	in.locknum = cpu_to_be32(lock);
	in.cid     = md->mycid;

	err = trans_io(md, (char*)&in, sizeof(dlock_cdb_t),
			(char*)&out, sizeof(dlock_rdf_t) );

	if( err !=0 || out.code != RDFdlock)
	printk("do_dlock: err = %d; out.code = %d\n", err, out.code);
	if(err) return err;
	if(out.code != RDFdlock) return -EIO;

	rtn->res = out.result;
	rtn->exp = out.expired;
	rtn->state   = out.state;
	rtn->vn  = be32_to_cpu(out.version_number);
	data->cidcnt = out.num_holders;

	return 0;
}

/******************************************************************************
 * do_refresh()
 * refreshes all of the locks held by cid.  This will be called from a
 * timer of some kind.
******************************************************************************/
int do_refreshAll(dlock_tcp_t *md)
{
	dlock_cdb_t in;
	dlock_rdf_t out;
	int err;

	memset(&in, 0, sizeof(dlock_cdb_t));
	memset(&out, 0, sizeof(dlock_rdf_t));

	in.code    = CDBdlock;
	in.action  = 0x04;
	in.ivn     = 0;
	in.locknum = ~0; /*Funny how this is the same in either byte order */
	in.cid     = md->mycid;

	err = trans_io(md, (char*)&in, sizeof(dlock_cdb_t),
			(char*)&out, sizeof(dlock_rdf_t) );

	if( err !=0 || out.code != RDFdlock)
	printe("do_refreshAll: err = %d; out.code = %d\n", err, out.code);
	if(err) return err;
	if(out.code != RDFdlock) return -EIO;

	return 0;
}

/******************************************************************************
 * do_modepage()
 * For getting the max number of dlocks & max number of cids on each lock 
******************************************************************************/
int do_modepage(dlock_tcp_t *md, u8 *maxclnt,
		u32 *numlock, u32 *lock_timeout)
{
	mode_page_t in, out;
	int err;

	memset(&in, 0, sizeof(mode_page_t));
	memset(&out, 0, sizeof(mode_page_t));

	in.code = CDBmoder;

	err = trans_io(md, (char*)&in, sizeof(mode_page_t),
			(char*)&out, sizeof(mode_page_t) );

	if( err !=0 || out.code != RDFmode)
	printe("do_modepage: err = %d; out.code = %d\n", err, out.code);
	if(err) return err;
	if(out.code != RDFmode) return -EIO;

	*maxclnt = out.max_clients;
	*numlock = be32_to_cpu(out.num_locks);
	*lock_timeout = be32_to_cpu(out.lock_time_ms);

	return 0;
}
