#include <sys/param.h>

#include <sys/systm.h>
#include <sys/kernel.h>

#include <sys/exec.h>
#include <sys/conf.h>
#include <sys/sysent.h>
#include <sys/lkm.h>
#include <sys/errno.h>
#include <sys/module.h>
#include <sys/namei.h>
#include <sys/fcntl.h>

#include <sys/buf.h>
#include <sys/sysproto.h>				    /* for sync(2) */
#include <sys/vnode.h>
#include <sys/disklabel.h>
#include <sys/dkbad.h>
/* #include <sys/lock.h> */

#define STATIC						    /* nothing while we're testing XXX */
#define BROKEN_GDB
#include "pool_fbsd.h"
#include "global.h"
#include "pool_debug.h"
#include "pool.h"



/* pool defines */

extern pool_t		pools[POOL_MAX_DEVICES];

static int pool_blocksizes[POOL_MAX_DEVICES];
static int pool_maxreadahead[POOL_MAX_DEVICES];
static int pool_size[POOL_MAX_DEVICES];  /* in KB */
static int print_buf(buf_t *);

static int pool_map(int minor, device_t **out_dev, u_long *rsector, u_long size);
int bdopen(device_t *dev, struct proc *p);
void bdclose(device_t *dev);
void pool_simpletest(void);


#if 0
spinlock_t	inet_dlock;	/* lock to access completion queue */
inetqueue_t	*head;
inetqueue_t	*tail;
#endif

/* Timer variables all const after init, no syncronization needed */
volatile void        *iotimer_addr;		/* physical address of timer */
uint64	 	      iotimer_max;
uint64                iotimer_rate_ps;          /* picoseconds per tick */
uint                  iotimer_size;


/*
 * declare your entry points as externs
 */

extern int  pool_open(dev_t dev, int oflags, int devtype, struct proc *p);
extern int  pool_close(dev_t dev, int fflag, int devtype, struct proc *p);
extern void pool_strategy(struct buf *bp);
extern int  pool_ioctl(dev_t dev, u_long cmd, caddr_t data,int fflag, struct proc *p);
extern int  pool_dump(dev_t dev);
extern int  pool_getsize(dev_t dev);
extern int  pool_read(dev_t dev, struct uio *uio, int ioflag);
extern int  pool_write(dev_t dev, struct uio *uio, int ioflag);
extern int  pool_poll(dev_t dev, int events, struct proc *p);

#if 0
extern int pool_load(struct lkm_table *lkmtp, int cmd);
extern int pool_unload(struct lkm_table *lkmtp, int cmd);
extern int pool_mod(struct lkm_table *lkmtp, int cmd, int ver);
#endif

extern void pool_attach(void *);
static int  poolinit(void);

/* taken from irix...  is it right? */
static void pool_done(struct buf *bp);
#if 0
extern void poolintr(struct clockframe *;)
#endif
#ifndef ACTUALLY_LKM_NOT_KERNEL
int pool_modevent(module_t mod, modeventtype_t type, void *unused);
#endif


#define BDEV_MAJOR 50
#define CDEV_MAJOR POOL_MAJOR

/*
 * build your device switch entry table
 */

static struct cdevsw pool_cdevsw = {
    (d_open_t *)      pool_open, /* open */
    (d_close_t *)     pool_close, /* close */
    (d_read_t *)      pool_read, /* read */
    (d_write_t *)     pool_write, /* write */
    (d_ioctl_t *)     pool_ioctl, /* ioctl */
    (d_stop_t *)      nostop, /* stop?? */
    (d_reset_t *)     noreset, /* reset */
    (d_devtotty_t *)  nodevtotty, /* ttys */
    (d_poll_t *)      pool_poll, /* select */
    (d_mmap_t *)      nommap, /* mmap */
    (d_strategy_t *)  pool_strategy, /* strategy */
	(char *)          "pool", /* d_name */
	(void *)          NULL,   /* d_spare */
	(int)             -1,     /* d_maj */
	(d_dump_t *)      pool_dump,       /* d_dump */
	(d_psize_t *)     pool_getsize,        /* d_psize */
	(u_int)           D_DISK,        /* d_flags */
	(int)             0,        /* d_maxio */
	(int)             -1        /* d_bmaj */
};

#ifdef ACTUALLY_LKM_NOT_KERNEL				    /* stuff for LKMs */
MOD_MISC(pool);


STATIC int 
pool_load(struct lkm_table *lkmtp, int cmd)
{
     BROKEN_GDB;
	pool_attach(NULL);
	printf("Dude loading pool\n");
    return 0;               /* OK */
}



/*
 * Function called when unloading the driver.
 */
STATIC int 
pool_unload(struct lkm_table *lkmtp, int cmd)
{
  struct sync_args dummyarg =	{0};
  BROKEN_GDB;
  if (pool_inactive()) {           /* is anything open? */

#if __FreeBSD__ < 3
	int retval;
#endif
	
	printf("pool: unloaded\n");
#if __FreeBSD__ < 3
	sync(curproc, &dummyarg, &retval);        /* write out buffers */
#else
	sync(curproc, &dummyarg);         /* write out buffers */
#endif
	/*free_vinum(0);*/              /* no: clean up */
#if __FreeBSD__ < 3
	bdevsw[BDEV_MAJOR] = NULL;          /* clear bdevsw */
#endif
	cdevsw[CDEV_MAJOR] = NULL;          /* and cdevsw */

	return 0;
  } else
	return EBUSY;
  return 0;
}

/*
 * Dispatcher function for the module (load/unload/stat).
 */
int 
pool_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
  BROKEN_GDB;
  MOD_DISPATCH(pool,             /* module name */
			   lkmtp,                /* LKM table */
			   cmd,                /* command */
			   ver,
			   pool_load,             /* load with this function */
			   pool_unload,             /* and unload with this */
			   lkm_nullcmd);
}

#else /* NOT LKM */

STATIC int 
pool_modevent(module_t mod, modeventtype_t type, void *unused)
{  
  struct sync_args dummyarg = {0};

  int rc;
  
  BROKEN_GDB;
  switch (type) {
  case MOD_LOAD:
	/*	  pool_attach(NULL); */
	rc = poolinit();
	printf("poolinit ret: %d\n",rc);
	return 0;					    /* OK */
  case MOD_UNLOAD:	
	
	sync(curproc, &dummyarg);			    /* write out buffers */
	printf("sync\n");
	Free(pools[1].p_spools->sp_devlist);
	Free(pools[1].p_spools);
	printf("Free\n");
	close_devices(&pools[1]);
	printf("closed devices going home\n");
	cdevsw[CDEV_MAJOR] = NULL;			    /* and cdevsw */
	return 0;
  default:
	break;
  }
  return 0;
}
moduledata_t pool_mod =
{
    "pool",
    (modeventhand_t) pool_modevent,
    0
};

DECLARE_MODULE(pool, pool_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
#endif

void
pool_attach(void *dummy)
{
  /*    BROKEN_GDB; */
  /* char *buf; */						    /* pointer to temporary buffer */
  struct _ioctl_reply *ioctl_reply;			    /* struct to return */
  
  printf("pool: loaded\n");
  
  /* We don't have a p pointer here, so take it from curproc */
  /*    myproc = curproc; */
  
  cdevsw_add_generic(BDEV_MAJOR, CDEV_MAJOR, &pool_cdevsw);
  
#ifdef DEVFS
#error DEVFS not finished yet
#endif
  
  ioctl_reply = NULL;					    /* no reply on longjmp */
}


/******************************************************************************
*******************************************************************************
** NAME
**	poolopen
**
** DESCRIPTION
**	This gets called for every open.  We basically just ensure the call is
**	to a valid minor number and increment the open counts accordingly.  In
**	addition, if this is the first open on the device, all the lower level
**	device comprising the pool are sent open calls.  (This ensures once a
**	user has the pool open lower level driver modules are not unloaded.)  A
**	special case is the opening of minor device 0.  This is the poolcomm
**	device and is the only pool_t that accepts assemble and remvove ioctls.
**	It is the only minor number that always exists and is not dynamically
**	assigned.  In addtion, access to it is serialized at the open level.
**	(That means we don't let more than two processes open it at the same
**	time.)  This prevents multiple copies of passemble from interfering.
**	Note that I believe it is still possible for a process to open poolcomm
**	and fork.  This essentially bypasses the precaution taken here.  DO NOT
**	do this.
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/


int
pool_open(dev_t dev, int oflags, int devtype, struct proc *p){

  int 		minor = minor(dev);
  pool_t	*pp;
  int		pl;
  
  printf("pool_open\n");
  
  IFDEBUG(POOL_DEBUG1,
	  printf("poolopen: minor=%d\n", minor););

  if (minor >= POOL_MAX_DEVICES)  {
    printf("POOL--poolopen: minor number out of range - %d\n", minor);
    return KRET(ENXIO);
  }

  pp = &pools[minor];
  printf("pool_open pp 0x%x\n",pp);
  print_pool(pp);
  /* if this is an open on the comm device handle it specially.  Access to
   * the comm device is exclusive */

  if (minor == 0) { 
    POOL_SLOCK(pl, &pp->p_mutex, pldisk);

    /* make sure the comm device exists */
    ASSERT((pp->p_flags & PF_FREE) == 0);

    /* check to see if device is busy */
    if (!(pp->p_flags & PF_VALID)) {
	  POOL_SUNLOCK(&pp->p_mutex, pl);
	  return KRET(EBUSY);
	}
    if (pp->p_oblk > 0 ) {
      POOL_SUNLOCK(&pp->p_mutex, pl);
      return KRET(EBUSY);
    }

    pp->p_oblk++;
    
    POOL_SUNLOCK(&pp->p_mutex, pl);
	/*     MOD_INC_USE_COUNT; what does this do ? */
    return(0);
  }

  /* regular open */

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN))  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(ENXIO);
  }
  if (!(pp->p_flags & PF_VALID))
  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(EBUSY);
  }

  /* verify that all pointers are valid */
#if 0
  ASSERT(pp->p_subpoollist != NULL);
  ASSERT(pp->p_dlockdevs != NULL);
#endif
  /* set the proper flags and return */

  pp->p_oblk++;

  POOL_SUNLOCK(&pp->p_mutex, pl);
  /*   MOD_INC_USE_COUNT; what does it do */
  return(0);
}

int
pool_close(dev_t dev, int fflag, int devtype, struct proc *p){

  int 		minor = minor(dev);
  pool_t	*pp;
  /*   subpool_t *spoolp; */
  int		pl;

  IFDEBUG(POOL_DEBUG1,
	  printf("pool_close: minor=%d\n", minor););

  pp = &pools[minor];

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  ASSERT((pp->p_flags & PF_VALID) != 0);
  ASSERT((pp->p_flags & PF_FREE) == 0);
  ASSERT(pp->p_oblk > 0);

  pp->p_oblk--;

  POOL_SUNLOCK(&pp->p_mutex, pl);
  /*  MOD_DEC_USE_COUNT; */
  return(0);
}

void
pool_strategy(struct buf *ubp){

  struct buf		*lbp	= NULL;
  minor_t	 minor	= minor(ubp->b_dev);
  pool_t	*pp =  &pools[minor];
  subpool_t   *spoolp = NULL;
  device_t  *devp = NULL;
  pool_queue_t  *qp = NULL;

  /*  kdev_t    dev; */  /*&(&(pp->p_spools)[0].sp_devlist)[0]->d_dev; */
  int s;
  int ret;
  /*  kdev_t    rdev; */
  u_long  rsector;

  pp =  &pools[minor];                                 
  spoolp = &(pp->p_spools[0]);
/*   devp =  &(spoolp->sp_devlist[0]);  */

  

#if 0 
  IFDEBUG(POOL_DEBUG1,
		  printf("Entered poolstrategy()\n");
		  printf("  Buffer pointer: 0x%08x\n", ubp);
		  printf("  Device: (%d) %d/%d\n", 
				 ubp->b_dev, major(ubp->b_dev), minor(ubp->b_dev));
		  printf(" Request: %s  offset 0x%x, length %ld resid %ld\n",
				 ubp->b_flags & B_READ ? "Read" : "Write",
				 ubp->b_blkno,
				 ubp->b_bcount,
				 ubp->b_resid);
		  printf("Dev %d\n",dev););
#endif
#if 0
  printf(" Request: %s  offset 0x%x(%d), length %ld resid %ld\n",
		 ubp->b_flags & B_READ ? "Read" : "Write",
		 ubp->b_blkno,
		 ubp->b_blkno,
		 ubp->b_bcount,
		 ubp->b_resid);
#endif
  
  rsector = ubp->b_blkno;
  ret = pool_map(minor,&devp /*out*/,
				 &rsector/*in/out*/ ,
				 ubp->b_bcount);
  if (ret){
	printf("pool_map: error %d\n",ret);
	ubp->b_error = ret;
	ubp->b_flags |= B_ERROR;
	biodone(ubp);
	return ;
  }
  /*  devp =  &(spoolp->sp_devlist[rdev]); */
  
  if (lbp == NULL) {
	lbp = (buf_t *)Malloc(sizeof(buf_t));
  }
  qp = (pool_queue_t *)Malloc(sizeof(pool_queue_t));
  qp->q_pp = pp;
  qp->q_ubp = ubp;

  bcopy(ubp,lbp,sizeof(struct buf));
  lbp->b_dev = devp->d_dev;
  lbp->b_vp = devp->vp;
  lbp->b_blkno = rsector;

  lbp->b_flags = lbp->b_flags | B_CALL;
  lbp->b_iodone = (void (*)(struct buf *))pool_done;
  lbp->b_driver1 = (void *)qp;

  if ((lbp->b_flags & B_READ) == 0)			    /* write */
	lbp->b_vp->v_numoutput++;		


  /*    print_buf(lbp); */
  s = splbio();
  (*bdevsw[major(lbp->b_dev)]->d_strategy) (lbp);
  splx(s);
  
}


/******************************************************************************
 ******************************************************************************
 **
 ** void pool_done()
 **
 ** Description:
 **   This routine services all q_lower requests, then calls biodone, and then
 **   cleans up. This routine serves as the functional equivalent of an inter-
 **   rupt handler for the driver.
 **
 ** Input(s):
 **  *bp - Pointer to the buffer structure.
 **
 ** Output(s):
 **
 **
 ******************************************************************************
 *****************************************************************************/

static void 
pool_done(buf_t *bp)
{
  pool_queue_t	*qp;
  buf_t		*ubp;
  buf_t		*lbp, *lbp_tmp;
  int		pl;	/* for LOCK */
#if 0
  int		pl2;	/* for LOCK */
  uint64	start, stop, usec, diff;
  pool_t	*pp;
  uint32       	b_flags, b_xflags, b_bcount;
  int		i;
#endif

#if 0
  IFDEBUG(POOL_DEBUG1,
		  printf("pool_done: "););
#endif
  
  qp = bp->b_driver1;
  ASSERT(qp);

  /* could have multiple calls on this pool queue, so lock */
#if 0
  POOL_SLOCK(pl,&qp->q_mutex, pldisk);

  qp->q_donecnt++;
#endif
  ubp = qp->q_ubp;
#if 0
  IFDEBUG(POOL_DEBUG1,
		  printf("bp=0x%lx  ubp=0x%lx\n",(void *)bp,(void *)ubp););
#endif
  biodone(ubp);
  return;
  /* XXX fix this. Perform some error checking */

  if (bp->b_error)
  {
    printf("pool_done: error - %d\n", bp->b_error);
    print_buf(bp);
    ubp->b_error = bp->b_error;
	ubp->b_flags |= B_ERROR;
  }

  IFDEBUG(POOL_DEBUG_BUFT,
	  print_buf(bp););

#if 0
  start = qp->q_start;
  pp = qp->q_pp;
#endif

#if 0
  POOL_SLOCK(pl2, &pp->p_statslock, pldisk);

  for (i=0; i<POOL_DSTATS_MAX; i++)
  {
    IFDEBUG(POOL_DEBUG2,
	    printf("pool_done: i=%d, b_dev=%d, s_dev=%d\n", 
		   i, bp->b_dev, (pp->p_dstats)[i].s_dev););
	
    if (bp->b_dev == (pp->p_dstats)[i].s_dev)
      break;
  }

  if (i == POOL_DSTATS_MAX)
  {
    printf("pool_done: device not found (%d,%d)\n", 
	   emajor(bp->b_dev), minor(bp->b_dev));
  }
  else
  {
#if 0
    if (bp->b_flags & B_READ)
    {
      if ((pp->p_dstats)[i].s_out_reads > 0)
	(pp->p_dstats)[i].s_out_reads--;
    }
    else
    {
      if ((pp->p_dstats)[i].s_out_writes > 0)
	(pp->p_dstats)[i].s_out_writes--;
    }
#endif
  }
  POOL_SUNLOCK(&pp->p_statslock, pl2);
#endif

  /* Transfer amount -- scary */

  ubp->b_resid += bp->b_resid;

  /* Still more to come */

  if (qp->q_donecnt < qp->q_lower)
  {
    IFDEBUG(POOL_DEBUG2,
	    printf("  More to come...\n"););
    POOL_SUNLOCK(&qp->q_mutex, pl);
    return;
  }

  /* Free buffers */

  lbp = qp->q_lbp;

  while (lbp != 0)
  {
    lbp_tmp = lbp;
#if 0
    lbp = lbp->b_forw;
#endif
#if 0
    kmem_free(lbp_tmp, sizeof(buf_t));
#elsif 9
    freerbuf(lbp_tmp);
#endif
  }

  /* Remove a queue element from the queue */

  POOL_SUNLOCK(&qp->q_mutex, pl);
#if 0
  kmem_free(qp, sizeof(pool_queue_t));
#endif

#if 0
  /* Save the statistics data before releasing ubp */

  b_flags = ubp->b_flags;
  b_xflags = ubp->b_xflags;
  b_bcount = ubp->b_bcount;
#endif

  /* Biodone - errors and counts should be included */
  printf("pool_done... biodone\n");
  biodone(ubp);

#if 0
  /* Stop timing, calculate microseconds */

  TIMER(stop);

  if (stop < start) {
      diff = ((uint64)(iotimer_max - start)) + 1 + stop;
    } else {
      diff = stop - start;
    }
      
  usec = (diff*iotimer_rate_ps)/1000000;

  IFDEBUG(POOL_DEBUG_TIME,
	  printf("diff  = %lld\n", stop-start);
	  printf("start = %lld\n", start);
	  printf("stop  = %lld\n", stop);
	  printf("diff  = %lld\n", diff);
	  printf("usecs  = %lld\n", usec););

  /* Update statistics */

  POOL_SLOCK(pl,&pp->p_statslock, pldisk);

  if (b_flags & B_READ) {
    pp->p_stats.s_reads++;
    pp->p_stats.s_readblks += b_bcount/DEV_BSIZE;
    pp->p_stats.s_readtime += usec;

    if (b_xflags & B_GFS_INVMISS) {
      IFDEBUG(POOL_DEBUG2,
	      printf("  Invalidated!\n"););

      pp->p_stats.s_invmiss++;
      pp->p_stats.s_invtime += usec;
    }
  } else {
    pp->p_stats.s_writes++;
    pp->p_stats.s_writeblks += b_bcount/DEV_BSIZE;
    pp->p_stats.s_writetime += usec;
  }
  POOL_SUNLOCK(&pp->p_statslock, pl);
#endif
}

int
pool_ioctl(dev_t dev, u_long cmd, caddr_t data,int fflag, struct proc *p){

  pool_t	*pp    	= NULL;
  int            rc 	= 0;
  int            minor 	= minor(dev);
  int		 pl    	= 0;			/* for LOCK */
#if 0
  int		 s     	= 0;
  int		 i;
  int		 len;
#endif

#if 0
  IFDEBUG(POOL_DEBUG1,
	  printf("Entered poolioctl(): ");
	  printf("  Device:  (%ud) %d/%d\n", dev,(int)major(dev), (int)minor(dev));
	  printf("  Command: 0x%x 0x%x\n", cmd,_POOLC_(cmd)););
#endif

  if (minor < 0 || minor >= POOL_MAX_DEVICES) {
    return KRET(ENXIO);
  }

  pp = &pools[minor];
#if 0
  POOL_SLOCK(pl,&pp->p_mutex, pldisk);
  if (pp->p_flags & PF_FREE) {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(ENXIO);
  }
  ASSERT(POOL_BUSY(pp));
  ASSERT(pp->p_flags & PF_VALID);
  POOL_SUNLOCK(&pp->p_mutex, pl);
#endif

  /* Do commands */

  switch(cmd) {
  case POOL_DEBUG_ON:  /*  Turn on debug bits.  */
  {
    uint64_t x;
    
    if (copyin((void *)data,&x,sizeof(uint64))) {
      return KRET(EFAULT);
    }
#ifdef DEBUG
    printf("Pool driver debugging enabled.\n");
    printf("  Flags: 0x%16qX + 0x%16qX = ", pool_dflags, x);
#endif
    pool_dflags |= x;
#ifdef DEBUG
    printf("0x%16qX\n", pool_dflags);
#endif
    break;
  }
  case POOL_DEBUG_OFF:  /*  Turn off debug bits.  */
  {
    uint64 x;
	  
    if (copyin((void *)data,&x,sizeof(uint64))) {
      return KRET(EFAULT);
    }

#ifdef DEBUG
    printf("Pool driver debugging disabled.\n");
    printf("  Flags:  0x%16qX - 0x%16qX = ", pool_dflags, x);
#endif
    pool_dflags &= ~x;
#ifdef DEBUG
    printf("0x%16qX\n", pool_dflags);
#endif
    break;
  }
  case POOL_ADD: 
  {
    pool_add_t	add;

    if (minor) {
      return KRET(EINVAL);
    }
    if (copyin(&add, (void *)data, sizeof(add)) != 0) {
      return KRET(EFAULT);
    }
    
    /* FIXME */
    /* minus one because the disk_partition_header is 1 BB after the label */
#if 0
    ASSERT( add.pa_label->pl_labelsize <= (POOL_SIZE_PHEADER(0) - 1) )
#endif
      

    rc = add_pool(&add);

#if 0
    pool_size[minor(add.pa_pool_dev)] = (add.pa_label->pl_capacity >> 1);
#endif

    if (copyout((void *)data, &add, sizeof(add)) != 0) {
      if (!rc)
	rc = KRET(EFAULT);
    }
    break;
  }
  case POOL_REMOVE:
  {
    uint32_t	dev;   /* minor number of device to remove */

    if (minor) { 
      return KRET(EINVAL);
    }
    if (copyin((void *)data, (void *)&dev, sizeof(dev)) != 0) {
      rc = KRET(EFAULT);
      break;
    }
    rc = remove_pool(dev);
    break;
  }
  case POOL_LIST:
  {
    pool_list_t	list;
    
    if (minor) {
      return KRET(EINVAL);
    }
    if (copyin((void *)data, (void *)&list, sizeof(list)) != 0) {
      return KRET(EFAULT);
    }

    rc = list_pool(&list);

    if (copyout((void *)&list, (void *)data, sizeof(list)) != 0) {
      if (!rc)
	rc = KRET(EFAULT);
    }
    break;
  }
  case POOL_SPLIST:
  {
    pool_splist_t list;

    if (copyin((void *)data, &list, sizeof(list)) != 0) {
      return KRET(EFAULT);
    }

    if (list.spl_count <= 0) {
      rc = KRET(EFAULT);
      break;
    }

    rc = list_spool(pp, &list);

      /* copy return values back out */

    if (copyout((void *)&list, (void *)data, sizeof(list)) != 0) {
      if (!rc)
	rc = KRET(EFAULT);
    }
    break;
  }
#if 0
  case POOL_IND_CACHE:
  {
    fsync_dev(dev);
    invalidate_buffers(dev);
    break;
  }
#endif
  case DIOCGDINFO:
	printf(" DIOCGDINFO\n");
	get_volume_label(pp, (struct disklabel *) data);
	break;
  case DIOCSDINFO:
	printf(" DIOCSDINFO\n");
	break;
  case DIOCWDINFO:
	printf(" DIOCWDINFO\n");
	break;
  case DIOCGPART:{
	get_volume_label(pp, &pp->p_label);
	((struct partinfo *) data)->disklab = &pp->p_label;
	((struct partinfo *) data)->part = &pp->p_label.d_partitions[0];
	printf(" DIOCGPART\n");
	break;
  }
  case DIOCRFORMAT:
	printf(" DIOCRFORMAT\n");
	break;
  case DIOCWFORMAT:
	printf(" DIOCWFORMAT\n");
	break;
  case DIOCSSTEP:
	printf(" DIOCSSTEP\n");
	break;
  case DIOCSRETRIES:
	printf(" DIOCSRETRIES\n");
	break;
  case DIOCWLABEL:
	printf(" DIOCWLABEL\n");
	break;
  case DIOCSBAD:
	printf(" DIOCSBAD\n");
	break;
  case DIOCSBADSCAN:
	printf(" DIOCSBADSCAN\n");
	break;
	
  default:
    rc = KRET(EINVAL);
  }
  return (rc);
}

int
pool_dump(dev_t dev){
  printf("pool_dump\n");
  return 0;
}

int
pool_getsize(dev_t dev){
  printf("pool_getsize\n");
  return 0;
}

int
pool_read(dev_t dev, struct uio *uio, int ioflag){
  /*  printf("pool_read\n"); */
  if (minor(dev) == 0) {
      printf("poolread: Can't write to poolcomm\n");
      return (EPERM);
    }
  if (minor(dev) > POOL_MAX_DEVICES)
    return -ENXIO;

  return(physio(pool_strategy, NULL, dev, B_READ, minphys, uio));
}

int
pool_write(dev_t dev, struct uio *uio, int ioflag){
/*   printf("pool_write\n"); */
 
  if (minor(dev) > POOL_MAX_DEVICES)
    return -ENXIO;
 
  if (minor(dev) == 0) {
	printf("poolwrite: Can't write to poolcomm\n");
	return (EPERM);
  }
  
  return physio(pool_strategy, 0, dev, B_WRITE, minphys, uio);
}

int
pool_poll(dev_t dev, int events, struct proc *p){
  printf("pool_poll\n");
  return 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	timestamp
**
** DESCRIPTION
**     
**
*******************************************************************************
******************************************************************************/

time_t
timestamp(void)
{
  struct timeval tv;

#if defined(linux)
  do_gettimeofday(&tv);
#elsif defined(__FreeBSD__)
  getmicrotime(&tv);
#endif

  return( tv.tv_sec );
}


static int 
poolinit(void){
  int i;

  printf("POOL v%d.%d.%d: %s at %s installed\n", POOL_VERSION >> 16, 
         (POOL_VERSION >> 8) & 0xFF, (POOL_VERSION) & 0xFF, __DATE__,
         __TIME__);

#if 0
  if (register_blkdev (POOL_MAJOR, "pool", &pool_fops)) {
      printf ("Unable to get major %d for pool\n", POOL_MAJOR);
      return (-1);
    }
#endif

  cdevsw_add_generic(BDEV_MAJOR, CDEV_MAJOR, &pool_cdevsw);

#if 0
  blk_dev[POOL_MAJOR].request_fn=do_pool_request;
  blk_dev[POOL_MAJOR].makerq_fn=NULL;
  blk_dev[POOL_MAJOR].map_fn=poolmap;
  blk_dev[POOL_MAJOR].current_request=NULL;
  read_ahead[POOL_MAJOR]=INT_MAX;
#endif

  for (i=0; i < POOL_MAX_DEVICES; i++) {
    pool_blocksizes[i] = 512;
    pool_maxreadahead[i] = 0;
    pool_size[i] = 0;
  }

#if 0
  blksize_size[POOL_MAJOR]=pool_blocksizes;
  blk_size[POOL_MAJOR]=pool_size;
  max_readahead[POOL_MAJOR]=pool_maxreadahead;
#endif

  /* initialize pools[] array */
  memset(pools, 0, POOL_MAX_DEVICES * sizeof(pool_t));  
  for (i=0; i < POOL_MAX_DEVICES; i++) {
    simple_lock_init(&pools[i].p_mutex);
    simple_lock_init(&pools[i].p_statslock);
    simple_lock_init(&pools[i].p_dlmutex);
    pools[i].p_flags = PF_VALID | PF_FREE;
  }
  
  pools[0].p_flags = PF_VALID;

  /* Initialize the debugging flags  */
  
  /*  pool_dflags = (0xFFFFFFFFFFFFFFFFll); */
  pool_dflags = (0x0ll);
  pool_dflags = POOL_DEBUG_MAP | POOL_DEBUG_OPEN | POOL_DEBUG1;
  
  /* Simple setup for developemnt */
  pool_simpletest();

  return (0);
}

/******************************************************************************
*******************************************************************************
** NAME
**	print_buf
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

static int
print_buf(buf_t *bp) {
  printf("hash      0x%x\n", bp->b_hash);
  printf("vnbufs    0x%x\n", bp->b_vnbufs);
  printf("bfreelist 0x%x\n", bp->b_freelist);
  printf("b_act     0x%x\n", bp->b_act);
  printf("b_proc    0x%x\n", bp->b_proc);
  printf("Flags:	0x%x\n", bp->b_flags);
  printf("qindex:	  %d\n", bp->b_qindex);
  printf("usecount:   %d\n", bp->b_usecount);
  printf("xFlags:	0x%x\n", bp->b_xflags);
  printf("Error:	  %d\n", bp->b_error);
  printf("bufsise     %u\n", bp->b_bufsize);
  printf("Xfer Cnt    %lu\n", bp->b_bcount);
  printf("Residual    %lu\n", bp->b_resid);
  printf("Device:	  %d (%d,%d)\n",bp->b_dev, major(bp->b_dev), minor(bp->b_dev));
  printf("data      0x%x\n",bp->b_data);
  printf("kvabase   0x%x\n",bp->b_kvabase);
  printf("kvasize   0x%x\n",bp->b_kvasize);
  printf("lblkno      %d\n",bp->b_lblkno);
  printf("blkno       %d\n",bp->b_blkno);
  printf("offset      %d\n",bp->b_offset);
  printf("iodone    0x%x\n",bp->b_iodone);
  printf("iodone_ch 0x%x\n",bp->b_iodone_chain);
  printf("vp        0x%x\n",bp->b_vp);
  printf("dirtyoff    %d\n",bp->b_dirtyoff);
  printf("dirtyend    %d\n",bp->b_dirtyend);
  printf("b_rcred   0x%x\n",bp->b_rcred);
  printf("b_wcred   0x%x\n",bp->b_wcred);
  printf("validoff    %d\n",bp->b_validoff);
  printf("validend    %d\n",bp->b_validend);
  printf("pblkno    0x%x\n", bp->b_pblkno);
  printf("saveaddr  0x%x\n", bp->b_saveaddr);
  printf("savekva   0x%x\n", bp->b_savekva);
  printf("driver1   0x%x\n", bp->b_driver1);
  printf("driver2   0x%x\n", bp->b_driver2);
  return (0);
}

void
pool_simpletest(void){

  int which=1;
  int ret;
  pool_t     		*new_pp = NULL;

  subpool_t *spool;
  device_t *onedev;
  device_t *twodev;
  
  new_pp = &pools[which];
   
  pools[which].p_spools = spool = Malloc(sizeof(subpool_t));
  printf("simpltest: 0x%x\n",pools[which].p_spools);
  strcpy(pools[which].p_name,"bart");
  pools[which].p_flags =  (PF_VALID | PF_LOWEROPEN);

  spool->sp_start = 0;
/*   spool->sp_len = 16772544; */
   spool->sp_len = 40000;
  spool->sp_devs = 2;
  spool->sp_stripe = 32;

  pools[which].p_capacity = spool->sp_len;
  pools[which].p_num_spools = 1;

  spool->sp_devlist = Malloc(spool->sp_devs * sizeof(device_t)); 
  printf("simpltest: 0x%x\n",&(spool->sp_devlist[0]),&(spool->sp_devlist[1]));

  onedev = &(spool->sp_devlist[0]);
  onedev->d_start = 0;
  /*  onedev->d_nblks = 8386272; */
  onedev->d_nblks = 20000;
  onedev->d_dev = makedev(4,4);
  strcpy(onedev->devicename,"/dev/da0e");
  twodev = &(spool->sp_devlist[1]);
  /*  twodev->d_start = 8386272;
	  twodev->d_nblks = 8386272; */
  twodev->d_start = 20000;
  twodev->d_nblks = 40000;

  twodev->d_dev = makedev(4,12);
  strcpy(twodev->devicename,"/dev/da1e");
  printf("pool_simpletest pp 0x%x\n",&pools[which]);
  
  ret = open_devices(new_pp);

  if (ret != 0) {
/*     POOL_SLOCK(pl, &new_pp->p_mutex, pldisk); */
    new_pp->p_flags |= PF_VALID | PF_FREE;
/* 	POOL_SUNLOCK(&new_pp->p_mutex, pl);  */
    printf("pool_simpletest %d\n",ret);
  }
}

/******************************************************************************
*******************************************************************************
** NAME
**	bdopen
**
** DESCRIPTION
**	Open a lower level block device...
**
** RETURN VALUE
**	Unix errno
**
*******************************************************************************
******************************************************************************/

int 
bdopen(device_t *dev, struct proc *p){
  int error;
  struct nameidata nd;

  IFDEBUG(POOL_DEBUG_OPEN,
	  printf("Entering bdopen\n"););
  NDINIT(&nd, LOOKUP, FOLLOW, UIO_SYSSPACE, dev->devicename, p);
  error = vn_open(&nd, FREAD | FWRITE, 0);		    /* open the device */
  if (error != 0) {	
	printf("bdopen %s: failed with error %d\n", dev->devicename, error);
	return error;
  }
  dev->vp = nd.ni_vp;
  dev->p = p;
  printf("bdopen vp: 0x%x\n",dev->vp);	


  return (error);
}

void 
bdclose(device_t *dev) { 
  IFDEBUG(POOL_DEBUG_OPEN,
	  printf("Entering bdclose\n"););
  if (dev->vp) {
	vn_close(dev->vp, FREAD | FWRITE, &(dev->p)->p_ucred, dev->p);
	dev->vp->v_usecount = 0;
	dev->vp = NULL;
  }
  IFDEBUG(POOL_DEBUG_OPEN,
	  printf("Leaving bdclose\n"););
  
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_map
**
** DESCRIPTION   
**	rsector and size are in basic blocks and are relative to the
**	logical pool device.  pool_blkmap() is called to translate the 
**	logical device and logical block into the actual device and block
**	for the request.
**
** RETURN VALUE
**	The parameters rdev and rsector are changed to reflect the mapping.
**
*******************************************************************************
******************************************************************************/

static int 
pool_map(int minor, 
		 device_t **out_dev,
		 unsigned long *rsector, 
		 unsigned long size){
  pool_t	*pp = NULL;
  device_t	*p_dev;
  int	       	i;
  int	       	pl;	/* for LOCK */
  int	       	bc;
  uint64      	dev_sector;


  IFDEBUG(POOL_DEBUG1,
	  printf("Entered pool_map()  minor=%d, rsector=%ld, size=%ld\n",
		 minor, *rsector, size););

  if (minor == 0) {  	/* handle writes to poolcomm */    
    printf("pool_map: can not write to poolcomm.\n");
    return KRET(EPERM);
  }

  pp = &pools[minor];

/*   POOL_SLOCK(pl, &pp->p_mutex, pldisk); */

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID) || 
      !(pp->p_flags & PF_LOWEROPEN)) {
/*     POOL_SUNLOCK(&pp->p_mutex, pl); */
    printf("pool_map: No such minor device %d\n", minor);
    return KRET(EIO);
  }

/*   POOL_SUNLOCK(&pp->p_mutex, pl); */


  /* Simple bound test */

  if ((*rsector + (size >> BBSHIFT) ) > pp->p_capacity) {
    printf("pool_map: Error with offset or size\n");
    printf("pool_map: blkno=%u, bcount=%u/%u p_capacity=%u\n", 
		   (uint32)*rsector, size, (size>> BBSHIFT), pp->p_capacity);
    return KRET(EIO);
  }

  /* Fill in the lower-level buffer structure */

/*   if (pool_blkmap(pp, *rsector, size << BBSHIFT, &dev_sector, &bc, &p_dev)) { */
  if (pool_blkmap(pp, *rsector, size, &dev_sector, &bc, &p_dev)) { 
    return KRET(EIO);
  }

  if (sizeof(unsigned long) <= 4 && dev_sector > UINT_MAX) {
    printf("poolmap: attempted access beyond 32 bit sector limit.\n");
    return KRET(EIO);
  }

  /*  *rdev = p_dev->d_dev; */
  *out_dev = p_dev;
  *rsector = (u_long)dev_sector;

  return(0);
}

/*
 * get_volume_label returns a label structure to lp, which
 * is allocated by the caller 
 */
/* Quick hack... eventually needs to be added to disklabel.h */
#define DTYPE_POOL             13             /* pool volume */
void 
get_volume_label(pool_t *pool, struct disklabel *lp)
{
  bzero(lp, sizeof(struct disklabel));
  
  strncpy(lp->d_typename, "pool", sizeof(lp->d_typename));
  lp->d_type = DTYPE_POOL;
  strncpy(lp->d_packname, pool->p_name, min(sizeof(lp->d_packname), sizeof(pool->p_name)));
  lp->d_rpm = 2;			    /* to keep them guessing */
  lp->d_interleave = 1;
  lp->d_flags = 0;
  
  /*
   * Fitting unto the vine, a vinum has a single
   *  track with all its sectors 
   */
  lp->d_secsize = DEV_BSIZE;				    /* bytes per sector */
  lp->d_nsectors = pool->p_capacity;				    /* data sectors per track */
  lp->d_ntracks = 1;					    /* tracks per cylinder */
  lp->d_ncylinders = 1;				    /* data cylinders per unit */
  lp->d_secpercyl = pool->p_capacity;			    /* data sectors per cylinder */
  lp->d_secperunit = pool->p_capacity;			    /* data sectors per unit */
  
  lp->d_bbsize = BBSIZE;
  lp->d_sbsize = SBSIZE;
  
  lp->d_magic = DISKMAGIC;
  lp->d_magic2 = DISKMAGIC;
  
  /*
   * Set up partitions a, b and c to be identical
   * and the size of the volume.  a is UFS, b is
   * swap, c is nothing 
   */
  lp->d_partitions[0].p_size = pool->p_capacity;
  lp->d_partitions[0].p_fsize = 1024;
  lp->d_partitions[0].p_fstype = FS_BSDFFS;		    /* FreeBSD File System :-) */
  lp->d_partitions[0].p_fsize = 1024;			    /* FS fragment size */
  lp->d_partitions[0].p_frag = 8;			    /* and fragments per block */
  lp->d_partitions[SWAP_PART].p_size = pool->p_capacity;
  lp->d_partitions[SWAP_PART].p_fstype = FS_SWAP;	    /* swap partition */
  lp->d_partitions[LABEL_PART].p_size = pool->p_capacity;
  lp->d_npartitions = LABEL_PART + 1;
  strncpy(lp->d_packname, pool->p_name, min(sizeof(lp->d_packname), sizeof(pool->p_name)));
  lp->d_checksum = dkcksum(lp);
}
