#include <sys/types.h>
#include <sys/lock.h>
#include <sys/param.h>
#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_extern.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <machine/limits.h>
#include <ufs/ffs/fs.h> /* for BBSIZE (boot block) and SBSIZE (super block) */


#define POOL_MAJOR              125 
/* eventually need something that actually */

typedef struct  simplelock lock_t;
typedef int minor_t;
typedef size_t kdev_t;
typedef struct buf buf_t;



#if defined(__FreeBSD__)
#define us_delay(x) DELAY(x)
#define KRET(x) x
#define POOL_BUSY(pp) ((pp)->p_oblk > 0)
#define memset(x, y, z) bzero(x, z)
#define KM_SLEEP  GFP_KERNEL
#define vmalloc(size) kmem_alloc(kernel_map,size)
#define GMAJOR  major
#define GMINOR  minor
#define MAJOR  major
#define MINOR  minor
#define makedevice makedev
#define vmem_free(pntr, len)   kmem_free(pntr,len) /* this will then call the km
em_free macro */
#define kmem_alloc(len, what)  kmem_alloc(kernel_map,len)
#define kmem_free(pntr, len)   kmem_free(kernel_map,pntr,len)
#define BBSHIFT 9
#ifndef ASSERT
#define ASSERT(x)  \
  if (!(x)) {   \
    printf("ASSERTION FAILED: %s\n", #x); \
  } 
#define bptoq(bp)       ((pool_queue_t *)((bp)->b_driver2))
#endif /* ASSERT */
#endif /* FreeBSD */

#if defined(__FreeBSD__)
#ifdef __ANSI_CPP__
#define POOL_SUNLOCK(lockp,pl) \
  {\
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tunlocking %s, file %s, line %d, val: %x\n", \
                   #lockp, __FILE__, __LINE__, *lockp.lock_data);); \
     simple_unlock(lockp); \
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tunlocked %s, val %x\n", #lockp, *lockp.lock_data);); \
  }
#else
#define POOL_SUNLOCK(lockp,pl) \
  {\
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tunlocking %s, file %s, line %d, val: %x\n", \
		   "lockp", __FILE__, __LINE__, *lockp.lock_data);); \
     simple_unlock(lockp);   \
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tunlocked %s, lock %d\n", "lockp", *lockp.lock_data);); \
  }
#endif /* ANSI Cpp */

#ifdef __ANSI_CPP__
#define POOL_SLOCK(plret,lockp,plset) \
  {\
     Ifdebug(POOL_DEBUG_SLOCK, \
	     printf("\tlocking %s, file %s, line %d, val: %x\n", #lockp,\
		   __FILE__, __LINE__, *lockp.lock_data);); \
     simple_lock(lockp);  \
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tlocked %s, val %x, old pl %x\n", #lockp, *lockp.lock_data, \
		   plret);); \
  }	   
#else
#define POOL_SLOCK(plret,lockp,plset) \
  {\
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tlocking %s, file %s, line %d, val: %x\n", "lockp",\
		   __FILE__, __LINE__, *lockp.lock_data);); \
     simple_lock(lockp);  \
     IFDEBUG(POOL_DEBUG_SLOCK, \
	     printf("\tlocked %s, val %x, old pl %x\n", #lockp, *lockp.lock_data, \
		   plret);); \
  }	   
#endif /* ANSI Cpp */
#endif /* FreeBSD */

#if defined(__FreeBSD__)
typedef struct pool_queue_s
{
  lock_t                 q_mutex;
  struct pool_s         *q_pp;          /* ptr to pool struct */
  minor_t                q_minor;       /* minor of dev for this req */
  buf_t                 *q_ubp;         /* upper buffer */
  u_int32_t                 q_lower;       /* # of lower buffers */ 
  u_int32_t                 q_donecnt;     /* # of lower requests done */ 
  buf_t                 *q_lbp;         /* lower buffer(s) */
  u_int64_t                 q_start;       /* start time */
} pool_queue_t;
#endif

#if defined(__FreeBSD__)
#define B_GFS_CACHE     0x00000040     /* device will cache blocks */
#define B_GFS_INVMISS   0x00000080     /* invalidation miss */
#define POOL_RESET_TIMEOUT_US   20000000
#define POOL_CNF_DELAY_US       50000
#endif

#define FMT_64 "q"
#define FMT_32 ""


#define NSPMajorVersion 3
#define NSPMinorVersion 0
#define NSPUpdateLevel  0

#ifdef REALLYKERNEL
#ifdef VINUMDEBUG
#define Malloc(x)  MMalloc ((x), __FILE__, __LINE__)        /* show where we cam
e from */
#define Free(x)    FFree ((x), __FILE__, __LINE__)          /* show where we cam
e from */
caddr_t MMalloc (int size, char *, int);
void FFree (void *mem, char *, int);
#else
#define Malloc(x)  malloc((x), M_DEVBUF, M_WAITOK)
#define Free(x)    free((x), M_DEVBUF)
#endif
#else
#define Malloc(x)  malloc ((x))                             /* just the size */
#define Free(x)    free ((x))                               /* just the address 
*/
#endif
