#ifndef __PARTITION_OPPS_C__
#define __PARTITION_OPPS_C__

#include "partition_opps.h"
#include "print.h"

/*
  This file includes most of the serv'ers behavior. 
Most of these functions are named identically to their 
name in the dlock specification of the scsi dlock command. 
*/

/*******************************************************
 *All actions passed with the dlock command descriptor block
 *are handled here. This function merely checks to see if the
 *lock involved is expired, decodes the command field, and 
 *launches one of the functions bellow that actually deal
 *with it. 
 *******************************************************/
void part_actions(partition_locks *pal, 
		  dlock_cdb *cdb, 
		  dlock_rdf *rdf){
 /*  //int created=0; */


  if((cdb->lock_number >= pal->dlocks_per_partition) && 
     (cdb->lock_number!=-1)){
    bad_lock_number         (pal, cdb, rdf);
  }
  else{
    if(cdb->lock_number!=-1){
      deal_expired            (pal, cdb, rdf);
    }
    switch(cdb->action){
    case 0:
      no_operation          (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 1:
      lock_shared           (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 2:
      lock_exclusive        (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 3:
      force_lock_exclusive  (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 4:
      touch_lock            (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 5:
      unlock                (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 6:
      unlock_increment      (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 7:
      activity_on           (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    case 8:
      activity_off          (pal, cdb, rdf);
      all_actions           (pal, cdb, rdf);
      break;
    default:
      break;
    }
  }
/*   //dbprintf("\n"); */
/*   //if(created!=0){ */
/*   //dbprintf("memory leaking:%d\n", created); */
/*   //} */
  return;  
}
  

/****************************************************************
 *unique to mode sense and mode select commands
 ****************************************************************/
void  mode_actions       (partition_locks *pal, mode_cdb  *cdb,
			  mode_page_rdf *rdf){

  if(cdb->tag==0){
    rdf->max_clients=pal->max_lock_holds;
    rdf->num_locks=pal->dlocks_per_partition;
    rdf->lock_time_ms=pal->lock_timeout_interval;  
  }
  else{
    pal->max_lock_holds=cdb->max_clients;
    pal->dlocks_per_partition=cdb->num_locks;
    pal->lock_timeout_interval=cdb->lock_time_ms;
    rdf->max_clients=pal->max_lock_holds;
    rdf->num_locks=pal->dlocks_per_partition;
    rdf->lock_time_ms=pal->lock_timeout_interval;
  }  
}

/****************************************************************
 *unique to report expired commands
 ****************************************************************/
void expired_actions     (partition_locks *pal, dlock_cdb *cdb, 
			  rexpired_rdf *rex){
  
  report_expired(pal, cdb, rex);
} 
 
/********************************************************
 *as the name implies. This one does nothing. 
 ********************************************************/
void no_operation(partition_locks *pal, 
		  dlock_cdb *cdb, dlock_rdf *rdf){
  rdf->result=1;}


/*******************************************************************
 * -If the lock is expired_exclusive lock exclusively so that the 
 *initiator can do any error correction.
 * -If the lock is unlocked && not expired_exclusive, share-lock it. 
 *Add the initiator's cid to the cid lis. reset the expire time. 
 * -If the lock is locked_shared && room in the cid list. Add 
 *the initiator's cid to the cid list. Reset the expire time. 
 * -If there is no room in the cid list then return a failure result.
 * -If the lock is locked_exclusive not held by the initiator, nil. 
 *a failure result.
 * -If locked exclusive held by this initiator, reset expire time 
 *and change to locked_shared.
 *******************************************************************/
void lock_shared(partition_locks *pal, 
		 dlock_cdb *cdb, dlock_rdf *rdf){

  if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){
    rdf->result=1;
    if(pal->sdl[cdb->lock_number]->expired==EXPIRED_E){
      pal->sdl[cdb->lock_number]->state=STATE_LOCKED_E;
    }
    else{
      pal->sdl[cdb->lock_number]->state=STATE_LOCKED_S;
    }
    add_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    sdl_reset(pal->sdl[cdb->lock_number]);
  }
    
  else if(pal->sdl[cdb->lock_number]->state==STATE_LOCKED_S){
    if(pal->sdl[cdb->lock_number]->num_holders<pal->max_lock_holds){
     /*  //dbprintf("add_second shared holder\n"); */
      rdf->result=1;
      add_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
      sdl_reset(pal->sdl[cdb->lock_number]);
    }
    else{rdf->result=0;}
  }  
  
  else{
    if(cid_in_dlock(pal->sdl[cdb->lock_number], cdb->cid)){
      rdf->result=1;
      pal->sdl[cdb->lock_number]->state=STATE_LOCKED_S;
      sdl_reset(pal->sdl[cdb->lock_number]);
    }
    else{
      rdf->result=0;
    }
  }
}


/*****************************************************************
 * -If lock is unlocked, lock it exclusively, add initiator's cid
 *to the cid list and reset the expire time.
 * -If lock is locked_shared && held by only this initiator, lock
 *exclusively and reset the expire time.
 * -If lock is locked shared by any other initiators, return failure.
 * -If lock is locked exclusively return failure.
 *****************************************************************/
void lock_exclusive(partition_locks *pal, 
		    dlock_cdb *cdb, dlock_rdf *rdf){
  if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){
    rdf->result=1;
    pal->sdl[cdb->lock_number]->state=STATE_LOCKED_E;
    add_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    sdl_reset(pal->sdl[cdb->lock_number]);
  }

  else if(pal->sdl[cdb->lock_number]->state==STATE_LOCKED_S){
    if(pal->sdl[cdb->lock_number]->num_holders==1 &&
	 cid_in_dlock(pal->sdl[cdb->lock_number], cdb->cid)) {
      rdf->result=1;
      pal->sdl[cdb->lock_number]->state=STATE_LOCKED_E;
      sdl_reset(pal->sdl[cdb->lock_number]);
    }

    else{
      rdf->result=0;
    }
  }
  else{
    rdf->result=0;
    
    
  }
}

/***********************************************************************
 * -If unlocked, add cid to list, lock exclusive, reset expire time. 
 * -If locked and input version is the same as the lock version
 *number, clear cid list, add cid, lock exclusively, reset the
 *the expire time, increment the version number.
 * -If locked but the input version number is not the same as
 *the lock version number return failure.
 ***********************************************************************/
void force_lock_exclusive(partition_locks *pal,
			  dlock_cdb *cdb, dlock_rdf *rdf){

  if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){
    rdf->result=1;
    pal->sdl[cdb->lock_number]->state=STATE_LOCKED_E;
    add_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    sdl_reset(pal->sdl[cdb->lock_number]);
  }
  
  else if(((pal->sdl[cdb->lock_number]->version_number << 8) >> 8)==
	  ((cdb->input_version_number << 8) >> 8)){
    rdf->result=1;
    pal->sdl[cdb->lock_number]->expired=pal->sdl[cdb->lock_number]->state;
    pal->sdl[cdb->lock_number]->state=STATE_LOCKED_E;
    sdl_clear_cid(pal->sdl[cdb->lock_number]);
    add_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    pal->sdl[cdb->lock_number]->version_number++;
    sdl_reset(pal->sdl[cdb->lock_number]);
  }

  else{
    rdf->result=0;
  }
}

/**************************************************************
 * -If unlocked return failure.
 * -If locked && the cid is in the list && lock number 
 *is 0xffffffff touch all locks held by that cid.
 * -If locked && the cid is in the list reset expire time.
 * -If locked && the cid is not in the list return failure.
 **************************************************************/
void touch_lock(partition_locks *pal, 
		dlock_cdb *cdb, dlock_rdf *rdf){

  int i;
  if(cdb->lock_number==0xffffffff){
    for(i=0; i<pal->dlocks_per_partition; i++){
      if(cid_in_dlock(pal->sdl[i], cdb->cid)){
	sdl_reset(pal->sdl[i]);
      }
    }
  }
  else{
    if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){
      rdf->result=0;}
    
    else if(cid_in_dlock(pal->sdl[cdb->lock_number], cdb->cid)){/*cid in list*/
      rdf->result=1;
      sdl_reset(pal->sdl[cdb->lock_number]);
    }
    
    else{
      rdf->result=0;
    }
  }
}


/*********************************************************
 * -If the lock is unlocked return failure.
 * -If locked && the cid is in the list, remove the cid from
 *the list, unlock if no more holders, increment version if
 *activity is on. 
 * -If cid not in list return failure.
 *********************************************************/
void unlock(partition_locks *pal, 
	    dlock_cdb *cdb, dlock_rdf *rdf){

  if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){rdf->result=0;}

  else if(cid_in_dlock(pal->sdl[cdb->lock_number], cdb->cid)){
    rdf->result=1;
    rem_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    if(pal->sdl[cdb->lock_number]->act==ACTIVITY_ON){
      pal->sdl[cdb->lock_number]->version_number++;}
    if(pal->sdl[cdb->lock_number]->num_holders==0){
      pal->sdl[cdb->lock_number]->state=STATE_UNLOCKED;
    }}

  else{
    rdf->result=0;}
}


/*******************************************************************
 * Behaves exactly like unlock, except that it will increment the
 * version number regardless of the activity bit's state. 
 *******************************************************************/
void unlock_increment(partition_locks *pal, 
		      dlock_cdb *cdb, dlock_rdf *rdf){
  
  if(pal->sdl[cdb->lock_number]->state==STATE_UNLOCKED){
    rdf->result=0;}
  
  else if(cid_in_dlock(pal->sdl[cdb->lock_number], cdb->cid)){
    rdf->result=1;
    rem_dlock_cid(pal->sdl[cdb->lock_number], cdb->cid);
    pal->sdl[cdb->lock_number]->version_number++;
    if(pal->sdl[cdb->lock_number]->num_holders==0){
      pal->sdl[cdb->lock_number]->state=STATE_UNLOCKED;
    }
  }
  else{
    rdf->result=0;}
}


/***********************************************************
 *set the activity bit of the lock.
 ***********************************************************/
void activity_on(partition_locks *pal, 
		 dlock_cdb *cdb, dlock_rdf *rdf){
  rdf->result=1;
  pal->sdl[cdb->lock_number]->act=ACTIVITY_ON;
 }



/***************************************************************
 *clear the activity bit of the lock. 
 ***************************************************************/
void activity_off(partition_locks *pal, 
		  dlock_cdb *cdb, dlock_rdf *rdf){
  rdf->result=1;
  pal->sdl[cdb->lock_number]->act=ACTIVITY_OFF;
  pal->sdl[cdb->lock_number]->version_number++;
}



/*****************************************************************
 *searches the entire list of locks and checks each one to see if 
 *it has expired. Sets exp to report all
 *of the expired locks. Also sets the expired field of the lock. If
 *not expired clears bits in lock and exp.
 *
 *broken still
 *****************************************************************/ 
void report_expired(partition_locks *pal, 
		    dlock_cdb *cdb, 
		    rexpired_rdf *exp){
 
  int i;
  time_t currenttime, *ctp;
  ctp=NULL;
  for(i=0; i<pal->dlocks_per_partition; i++){
    currenttime=time(ctp);
    if((pal->sdl[i]->expire_time < (int)currenttime)
       && (pal->sdl[i]->state != STATE_UNLOCKED)){
      pal->sdl[i]->expired=pal->sdl[i]->state;
      pal->sdl[i]->state=STATE_UNLOCKED;
      pal->sdl[i]->version_number++;
      sdl_clear_cid(pal->sdl[cdb->lock_number]);
    }
/*     //exp->return_length=(pal->dlocks_per_partition)/8; */
    if(pal->sdl[i]->expired!=EXPIRED_NOT){
/*       //re_rdf_set_exlock(exp, i); */
    }
    else{
/*       //re_rdf_clear_exlock(exp, i); */
    }
  }
}


/***************************************************************
 *checks the lock to see if it has expired. sets the expired bit
 *of the lock if it has. 
 ***************************************************************/
void deal_expired(partition_locks *pal, 
		  dlock_cdb *cdb, dlock_rdf *rdf){

  int i;
  time_t currenttime, *ctp;

  ctp=NULL;
  currenttime=time(ctp);
  i=cdb->lock_number;
  if((pal->sdl[cdb->lock_number]->expire_time < (int)currenttime)
     && (pal->sdl[cdb->lock_number]->state != STATE_UNLOCKED)){
    pal->sdl[cdb->lock_number]->expired=
      pal->sdl[cdb->lock_number]->state;
    pal->sdl[cdb->lock_number]->state=STATE_UNLOCKED;
    pal->sdl[cdb->lock_number]->version_number++;
    sdl_clear_cid(pal->sdl[cdb->lock_number]);
  } 
} 


/**************************************************************
 *every action does this. sets the return data based on the 
 *lock status. including returning all the cidames.
 **************************************************************/
void all_actions(partition_locks *pal, 
		 dlock_cdb *cdb, dlock_rdf *rdf){

  int i,j;
  if(cdb->lock_number==-1){
    rdf->act = 0;
    rdf->state = 0;
    rdf->version_number= 0;
    rdf->expired = EXPIRED_NOT;
    rdf->num_holders=0;
    clr_cid_entry(rdf);
  }/* //special case for touch lock.(above)regular (below) */

  else{
    rdf->act =    pal->sdl[cdb->lock_number]->act;
    rdf->state =  pal->sdl[cdb->lock_number]->state;
    rdf->version_number= pal->sdl[cdb->lock_number]->version_number;
    rdf->expired= pal->sdl[cdb->lock_number]->expired;
    clr_cid_entry(rdf);
    j=pal->sdl[cdb->lock_number]->num_holders;
    for(i=0;i<j; i++){
      add_cid_entry(rdf, get_cid_list(i,pal->sdl[cdb->lock_number])->cid);
      if((i==j)&&(get_cid_list(i,pal->sdl[cdb->lock_number])->next!=NULL)){
	dbprintf("***\n***\n***\n***fuck***\n***\n***\n***\n");
      }      
    }
  }
  return;
}
   
/***************************************************************
 *if the lock number is out of bounds.
 ***************************************************************/
void bad_lock_number(partition_locks *pal, 
		 dlock_cdb *cdb, dlock_rdf *rdf){
  rdf->result=0;
  rdf->act=0;
  rdf->expired=0;
  rdf->state=0;
  rdf->version_number=0;
  clr_cid_entry(rdf);
}



#endif /* __PARTITION_OPPS_C__ */


