/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Andrew Barry <barry@lcse.umn.edu>
** 
**    Module name: mode_page.c
**
**    Description: functions for modifying mode_page command and reply
**    data structures.
**
*******************************************************************************
******************************************************************************/


#ifndef __MODE_PAGE_C__
#define __MODE_PAGE_C__

#include "mode_page.h"

mode_page *new_mode_page_cdb(){
  mode_page *mscdb;
  mscdb=(mode_page *)malloc(sizeof(mode_page));
  mscdb->code=CDBMODER;
  mscdb->max_clients=64;
  mscdb->num_locks=1024;
  mscdb->lock_time_ms=5000;
  return(mscdb);
}

void delete_mode_page(mode_page *mp){
  free(mp);
}

mode_page *new_mode_page_rdf(){
  mode_page *rdf;
  rdf=(mode_page *)malloc(sizeof (mode_page));
  rdf->code=RDFMODE;
  rdf->num_locks=DLOCKS_PER_PARTITION;
  rdf->max_clients=MAX_LOCK_HOLDS;
  rdf->lock_time_ms=1000*EXPIRE_TIME;
  return(rdf);
}

rexpired_rdf *new_rexpired_rdf(int numlocks){
  rexpired_rdf *rerdf;
  int i;
  rerdf=(rexpired_rdf *)malloc(sizeof(rexpired_rdf));
  rerdf->return_length=numlocks;
  rerdf->exlock_block=(uint8 *)malloc(numlocks*sizeof(uint8));
  for(i=0; i<numlocks; i++){
    rerdf->exlock_block[i]=0;
  }
  return(rerdf);
}

void delete_rexpired_rdf(rexpired_rdf *rdf){
  free(rdf);
}

uint8 *get_exlock_block(rexpired_rdf *rdf, int index){
  uint8 *elb;

  elb=&(rdf->exlock_block[index]);
  return(elb);
}

void re_rdf_set_exlock(rexpired_rdf *rdf, int lock){
  int diveight, rem;
  uint8 *elb;
  diveight= lock / 8;
  rem=      lock % 8;
  elb=get_exlock_block(rdf, diveight);
  switch(rem){
  case 0:
    *elb=*elb | FIRSTBIT;
    break;
  case 1:
    *elb=*elb | SECONDBIT;
    break;
  case 2:
    *elb=*elb | THIRDBIT;
    break;
  case 3:
    *elb=*elb | FOURTHBIT;
    break;
  case 4:
    *elb=*elb | FIFTHBIT;
    break;
  case 5:
    *elb=*elb | SIXTHBIT;
    break;
  case 6:
    *elb=*elb | SEVENTHBIT;
    break;
  case 7:
    *elb=*elb | EIGHTHBIT;
    break;
  }
}

void re_rdf_clear_exlock(rexpired_rdf *rdf, int lock){
  int diveight, rem;
  uint8 *elb;

  diveight= lock / 8;
  rem=      lock % 8;
  elb=get_exlock_block(rdf, diveight);
  switch(rem){
  case 0:
    *elb=*elb & !FIRSTBIT;
    break;
  case 1:
    *elb=*elb & !SECONDBIT;
    break;
  case 2:
    *elb=*elb & !THIRDBIT;
    break;
  case 3:
    *elb=*elb & !FOURTHBIT;
    break;
  case 4:
    *elb=*elb & !FIFTHBIT;
    break;
  case 5:
    *elb=*elb & !SIXTHBIT;
    break;
  case 6:
    *elb=*elb & !SEVENTHBIT;
    break;
  case 7:
    *elb=*elb & !EIGHTHBIT;
    break;
  }
}

#endif /*__MODE_PAGE_C__*/


