/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
**
**    Module name: pool_ondisk.c
**
**    Description:  This module converts all pool structures that exist on
**      disk to internal processor formats.  Handles structure alignment
**      and endianess issues.  All disk structures must be multiples of
**      8 bytes.  See also GFS/src/fs/ondisk.c
**
**
*******************************************************************************
******************************************************************************/

#ifdef __KERNEL__
#include <asm/byteorder.h>
#include <linux/string.h>
#else
#include "gfs_endian.h"
#include <string.h>
#endif  /*  __KERNEL__  */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <paths.h>
#include <ctype.h>

#include "global.h"
#include "pool_ondisk.h"



#define CPIN_8(x, y, z) {memcpy((x), (y), (z)); (y) += (z);}
#define CPOUT_8(x, y, z) {memcpy((y), (x), (z)); (y) += (z);}
#define CPIN_16(x, y) {memcpy((x), (y), 2); (y) += 2; be16_to_cpus((x));}
#define CPOUT_16(x, y) {uint16 _cp = cpu_to_be16p(x); memcpy((y), &_cp, 2); (y) += 2;}
#define CPIN_32(x, y) {memcpy((x), (y), 4); (y) += 4; be32_to_cpus((x));}
#define CPOUT_32(x, y) {uint32 _cp = cpu_to_be32p(x); memcpy((y), &_cp, 4); (y) += 4;}
#define CPIN_64(x, y) {memcpy((x), (y), 8); (y) += 8; be64_to_cpus((x));}
#define CPOUT_64(x, y) {uint64 _cp = cpu_to_be64p(x); memcpy((y), &_cp, 8); (y) += 8;}



/******************************************************************************
*******************************************************************************
**
** pool_label_out()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/
 
void pool_label_out(pool_label_t *pl, char *buf)
{
  memset(buf, 0, SIZEOF_POOL_LABEL);

  CPOUT_64(&pl->pl_magic, buf);
  CPOUT_64(&pl->pl_pool_id, buf);
  CPOUT_32(&pl->pl_version, buf);
  CPOUT_32(&pl->pl_subpools, buf);
  CPOUT_32(&pl->pl_sp_id, buf);
  CPOUT_32(&pl->pl_sp_devs, buf);
  CPOUT_32(&pl->pl_sp_devid, buf);  
  CPOUT_64(&pl->pl_blocks, buf);
  CPOUT_32(&pl->pl_striping, buf);  
  CPOUT_32(&pl->pl_sp_dldevs, buf);
  CPOUT_32(&pl->pl_sp_dlid, buf);
  CPOUT_32(&pl->pl_dlocks, buf);
  CPOUT_32(&pl->pl_dl_stripe, buf);
  CPOUT_8(&pl->pl_pool_name, buf, POOL_NAME_SIZE);
}


/******************************************************************************
*******************************************************************************
**
** pool_label_in()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/
 
void pool_label_in(pool_label_t *pl, char *buf)
{
  CPIN_64(&pl->pl_magic, buf);
  CPIN_64(&pl->pl_pool_id, buf);
  CPIN_32(&pl->pl_version, buf);
  CPIN_32(&pl->pl_subpools, buf);
  CPIN_32(&pl->pl_sp_id, buf);
  CPIN_32(&pl->pl_sp_devs, buf);
  CPIN_32(&pl->pl_sp_devid, buf);  
  CPIN_64(&pl->pl_blocks, buf);
  CPIN_32(&pl->pl_striping, buf);
  CPIN_32(&pl->pl_sp_dldevs, buf);
  CPIN_32(&pl->pl_sp_dlid, buf);
  CPIN_32(&pl->pl_dlocks, buf);
  CPIN_32(&pl->pl_dl_stripe, buf);
  CPIN_8(&pl->pl_pool_name, buf, POOL_NAME_SIZE);
}















