#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "misc.h"
#include "structs.h"
#include "state.h"

#include "global.h"
#include "wrappers.h"
#include "dlock_interface.h"

#define ASSERT(x)							\
{ if (!(x)) {								\
    printf("***ASSERT***: %s failed on input line %d\n", #x, line_nr);	\
    exit(1);								\
}									\
}

void sigsegv(int signo) {
  printf("SEGFAULT: basic\n");
  exit(1);
}

int old_val(char *c, int old) {

  if (*c == '-')
    return old;
  else
    return atoic(c);
}

void usage() {
  printf("usage: basic <input file> <transport type> <device>\n");
}


uint64 dl_flags = DLF_DPRINTF | DLF_ERROR;

typedef struct input_s {
  dlock_data_t d;
  uint32 prevstate, diff;
} input_t;


int main(int argc, char **argv) {
  char buf[1024];
  char arg[10][256];
  FILE *fp;
  dlock_data_t test_pr;
  device_handle_t *handle;
  input_t cur, prev;
  int line_nr = 1, cmd_nr = 0, x;

  if (argc < 4) {
    usage();
    return 1;
  }

  for (x=4; x < argc; x++) {
    int y;

    if (!strncmp(argv[x], "d=", 2)) {

      for (y=2; argv[x][y]; y++) {

	if (argv[x][y] == 'n')
	  dl_flags = 0;

	if (argv[x][y] == 'f')
	  dl_flags |= DLF_FUNCTIONS;

	if (argv[x][y] == 'd')
	  dl_flags |= DLF_DPRINTF;

	if (argv[x][y] == 'e')
	  dl_flags |= DLF_ERROR;

      }
      
    }
  }

  suite_init(); /* initialize tables */

  if ((fp = fopen(argv[1], "r")) == NULL)
    die("can't open file\n");

  handle = dl_open(argv[2], argv[3]);

  ASSERT(handle);

  /* 0          1      2   3  4     5      6   7    8     9     */
  /* prevstate action cid l# ivn, diff state exp holder result */

  while (fgets(buf, 1024, fp)) {
    dlock_state_t st;
    int vals;
    
    dprintf("\n-------------------------------\n");
    dprintf("line %d: %s", line_nr, buf);

    vals = sscanf(buf, "%s %s %s %s %s %s %s %s %s %s", 
		  arg[0], arg[1], arg[2], arg[3], arg[4], arg[5], arg[6], arg[7], arg[8], arg[9]);

    
    if (*arg[0] == '#' || vals == -1) {
      line_nr++;
      continue;
    }

    cmd_nr++;

    if (*arg[0] != '*')
      cur.prevstate = table_val(&state_tab, arg[0], prev.prevstate);

    cur.d.action = table_val(&action_tab, arg[1], prev.d.action);
    cur.d.clientid = old_val(arg[2], prev.d.clientid);
    cur.d.locknum = old_val(arg[3], prev.d.locknum);

    if (!strcmp(arg[4], "ver")) {
      //dprintf("** in versoin\n");
      cur.d.inputversion = prev.d.version;
    }
    else
      cur.d.inputversion = old_val(arg[4], prev.d.inputversion);

    cur.diff = old_val(arg[5], prev.diff);
    cur.d.state = table_val(&state_tab, arg[6], prev.d.state);
    cur.d.expired = table_val(&expired_tab, arg[7], prev.d.expired);
    cur.d.holders = old_val(arg[8], prev.d.holders);
    cur.d.result = old_val(arg[9], prev.d.result);

    if (*arg[0] != '*') {
      uint32 clientid;

      memset(&st, 0, sizeof(st));
      st.state = cur.prevstate;
    
      if (cur.prevstate == LOCKED_SHARED || cur.prevstate == LOCKED_EXCL)
	clientid = cur.d.clientid + 0xdeadbeef;  /* hackish, something other than cur.d.clientid */
      else
	clientid = cur.d.clientid;

      if (goto_state(handle, &st, cur.d.locknum, clientid) == -1)
	die("can't goto_state\n");
    
    }


    memcpy(&test_pr, &cur.d, sizeof(test_pr));

    if (do_dlock(handle, &test_pr))
      die("do_dlock failed on test\n");

    if (cmd_nr > 1) 
      ASSERT( ((test_pr.version - prev.d.version) == cur.diff)  );
    

    ASSERT( test_pr.state   ==   cur.d.state );
    ASSERT( test_pr.expired ==   cur.d.expired  );
    ASSERT( test_pr.holders ==   cur.d.holders  );
    ASSERT( test_pr.result  ==   cur.d.result );


    memcpy(&cur.d, &test_pr, sizeof(test_pr));
    memcpy(&prev, &cur, sizeof(cur));
    line_nr++;
  }

  return 0;
}

    

    
